;*********************************************************************
; ISR_Timer_Clock
;*********************************************************************
; $Id: isr_rtc.s,v 1.98.2.2 2014/04/01 21:46:40 jsmcortina Exp $

; * Copyright 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
; * James Murray and Kenneth Culver
; *
; * This file is a part of MS2/Extra.
; *
; * Origin: Al Grippo
; * Major: Recode in ASM, new features. James Murray / Kenneth Culver
; * Majority: James Murray / Kenneth Culver
; *
; * You should have received a copy of the code LICENSE along with this source,
; * ask on the www.msextra.com forum if you did not.
; *

.sect .text
.globl ISR_Timer_Clock
             nolist               ;turn off listing
             include "ms2extrah.inc"
             list                 ;turn listing back on

;**************************************************************************
; "0.1ms" section periodic interrupt
; maintains clocks amongst other functions
; **************************************************************************

ISR_Timer_Clock:
;    // .128 ms clock interrupt - clear flag immediately to resume count
   movb    #128, CRGFLG ; clear RTI interrupt flag

;    // also generate 1.024 ms, .10035 sec and 1.0035 sec clocks
   ldd	    lmms+0x2 ; free running clock(.128 ms tics) good for ~ 110 hrs
   ldx	    lmms
   addd     #1
   bcc      lm3a
   inx
lm3a:
   std      lmms+0x2
   stx      lmms

;    mms++;    // in .128 ms tics - reset every 8 tics = 1.0024 ms
   inc      mms

    brclr   flagbyte0, #FLAGBYTE0_ENGLOG, noenglog
    jsr     do_englog

noenglog:

; sequential injection injector channel 1
   ldd      inj1_count
   beq      injector_2
   subd     #1
   std      inj1_count
   bne      injector_2
   ; start injection for channel 1
   STACK_SOFT
   call    INJ1
   UNSTACK_SOFT

injector_2:
; sequential injection injector channel 2
   ldd      inj2_count
   beq      injector_3
   subd     #1
   std      inj2_count
   bne      injector_3
   ; start injection for channel 2
   STACK_SOFT
   call    INJ2
   UNSTACK_SOFT

injector_3:
; sequential injection injector channel 3
   ldd      inj3_count
   beq      injector_4
   subd     #1
   std      inj3_count
   bne      injector_4
   ; start injection for channel 3
   STACK_SOFT
   call    INJ3
   UNSTACK_SOFT

injector_4:
; sequential injection injector channel 4
   ldd      inj4_count
   beq      map_sampling
   subd     #1
   std      inj4_count
   bne      map_sampling
   ; start injection for channel 4
   STACK_SOFT
   call    INJ4
   UNSTACK_SOFT

map_sampling:
;;    brclr   flagbyte20, #FLAGBYTE20_USE_MAP, bypassmap ; is MAP enabled at all?
    brclr	flagbyte0, #FLAGBYTE0_MAPLOG, no_ml
    ldx     ATD0DR0
    jsr     do_maplog1
no_ml:

    ldab    map_deadman
    cmpb    #4
    bhs     deadmap
   brset    outpc.engine,#0x1,norm_map_sample
map_nonwindow:
   ldx      ATD0DR0
   stx      map_temp
    clr     map_sum
    clr     map_sum+1
    stx     map_sum+2
    movw    #1,map_cnt
   bset     flagbyte3,#flagbyte3_samplemap
   bset     flagbyte11,#FLAGBYTE11_MAP_AVG_RDY
   clr      map_deadman ; acknowledge that we have taken a sample
bypassmap:
   bra      done_min 

deadmap:
   ; something went wrong with MAP sampling and samples not happening
   bset     outpc.status3, #4 ; STATUS3_MAPERROR
   bra      map_nonwindow

norm_map_sample:
   brclr    flash4.mapsample_opt,#MAPSAMPLE_OPT_USE_AVG,map_sample_lowest
   brclr    flagbyte11,#FLAGBYTE11_MAP_AVG_TRIG,continue_norm_map
   movw     map_temp_sum,map_sum
   movw     map_temp_sum+2,map_sum+2
   movw     map_temp_cnt,map_cnt
   movw     #0,map_temp_sum
   movw     #0,map_temp_sum+2
   movw     #0,map_temp_cnt
   bset     flagbyte11,#FLAGBYTE11_MAP_AVG_RDY
   bclr     flagbyte11,#FLAGBYTE11_MAP_AVG_TRIG
   clr      map_deadman ; acknowledge that we have taken a sample

continue_norm_map:
   ldd      ATD0DR0
   addd     map_temp_sum+2
   std      map_temp_sum+2
   ldd      map_temp_sum
   adcb     #0
   adca     #0
   std      map_temp_sum
   ldd     map_temp_cnt
   addd    #1
   std     map_temp_cnt
    bra     done_min

map_sample_lowest:
   ldd      map_start_countdown
   beq      cont_map_check

   ; MAP sampling time?
   subd     #1
   std      map_start_countdown

;   bne      done_mapsample
   bne      cont_map_check  ; allow window to be running while MAP start is counting down
   movw     map_window_set,map_window_countdown
   movw     #0xFFFF,map_temp
    bset    flagbyte15, #FLAGBYTE15_MAPWINDOW
cont_map_check:
   ldd      map_window_countdown
   beq      done_mapsample
   subd     #1
   std      map_window_countdown
   ldx      ATD0DR0
    cmpx    mapadc_thresh
    bhs     no_mt1
    bset    flagbyte1,#flagbyte1_trig2active  ; MAP phase detection less than threshold
no_mt1:
   cmpx     map_temp
   bhi      cont_map_check2
   stx      map_temp
    movb    tooth_no, mapmin_tooth_temp
cont_map_check2:
   cpd      #0
   bne      done_mapsample
    bclr    flagbyte15, #FLAGBYTE15_MAPWINDOW
   bset     flagbyte3,#flagbyte3_samplemap
   clr      map_deadman ; acknowledge that we have taken a sample
    movb    mapmin_tooth_temp, mapmin_tooth
    brclr   flash4.mapsample_opt, #3, done_min ; bypass

; 4 element ring buffer
    clra
    ldab    map_temps_cnt
    aslb
    tfr     d,y
    ldx     map_temp    ; lowest value found for this event
    stx     map_temps, y
    brclr   flash4.mapsample_opt, #2, scmp1

    cmpb    #6
    blo     cmc2a
    clr     map_temps_cnt
    bra     cmc2b
scmp1:
    cmpb    #2
    blo     cmc2a
    clr     map_temps_cnt
    bra     cmc2b

cmc2a:
    inc     map_temps_cnt
cmc2b:

;pick smallest value
    ldy     #map_temps
    ldd     2,y+
    emind   2,y+
    brclr   flash4.mapsample_opt, #2, str_min
    emind   2,y+
    emind   0,y
str_min:
    std     map_temp

done_min:
   brclr    flagbyte6,#0x2,done_mapsample
   brset    flash4.MAFOption,#0x30,sample_ATD0DR0
   brset    flash4.MAFOption,#0x10,sample_ATD0DR7
   movw     ATD0DR6,maf_temp
   bra      done_mapsample
sample_ATD0DR0:
   movw     ATD0DR0,maf_temp
   bra      done_mapsample
sample_ATD0DR7:
   movw     ATD0DR7,maf_temp

done_mapsample:
    brclr	flagbyte0, #FLAGBYTE0_MAFLOG, no_mfl7
    ldy     mafport
    ldx     0,x
    jsr     do_maplog1
no_mfl7:

; tps_ring
    ldab    tps_ring_cnt
    incb
    cmpb    #7
    bls     tpsrs
    clrb
tpsrs:
    stab    tps_ring_cnt
    lslb
    sex     b,x
    ldy     ATD0DR3
    sty     tps_ring,x
;end tps_ring

   brset    flagbyte1,#flagbyte1_tstmode,do_testmode
   jmp      ir_notest

; --------- test mode in here, normally skipped  --------
do_testmode:
; test mode
   ldaa     testmode_glob
   cmpa     #3
   beq      tst_pulsedio

;FP removed

   ldx      testcnt
   beq      tst_choose
   dex
   stx      testcnt
   bra      tst_inj_pwm
tst_choose:
   movw     ramdata.testint,testcnt

   bclr     TCTL2, #0x44  ; set outputs low
   bset     CFORC, #0x0a  ; force inj off. Ensures they turn off if mode changed

   ldaa     testmode_glob
   bita     #1
   bne      tst_coils
   bita     #2
   bne      tst_inj
   bra      ir_notest

tst_coils:
   movb     #0xff, maxdwl  
   ldaa     ramdata.testsel
    lsra
    lsra
    lsra
    lsra
    brclr   flash4.EngStroke, #2, tst_sc_norm ; not rotary
    ldab    flash4.no_cyl
    andb    #0x1f
    cba
    bhs     tst_sc_trailing ; upper coils are trailing

tst_sc_norm:
   cmpa     #0x01
   beq      tds2
   cmpa     #0x02
   beq      tds4
   cmpa     #0x03
   beq      tds8
   cmpa     #0x04
   beq      tds10
   cmpa     #0x05
   beq      tds20
   movb     #1,dwellsel
   movb     #1,coilsel
   bra      tdsdone
tds2:
   movb     #2,dwellsel
   movb     #2,coilsel
   bra      tdsdone
tds4:
   movb     #4,dwellsel
   movb     #4,coilsel
   bra      tdsdone
tds8:
   movb     #8,dwellsel
   movb     #8,coilsel
   bra      tdsdone
tds10:
   movb     #0x10,dwellsel
   movb     #0x10,coilsel
   bra      tdsdone
tds20:
   movb     #0x20,dwellsel
   movb     #0x20,coilsel
tdsdone:
   ldd      TCNT
   addd     #48 ; slight delay
   std      TC_dwl ; set up dwell O/C
   bset     TIE,#TFLG_dwl
   movb     #TFLG_dwl, TFLG1

   ldaa     ramdata.testdwell
   ldab     #170
   mul
   addd     #48
   addd     TCNT
   std      TC_ign  ; set up spark O/C
   bset     TIE,#TFLG_ign
   movb     #TFLG_ign, TFLG1
   bra      ir_notest

tst_sc_trailing:
   cmpa     #0x02
   beq      tdst4
   cmpa     #0x03
   beq      tdst8
   cmpa     #0x04
   beq      tdst10
   cmpa     #0x05
   beq      tdst20
   bra      ir_notest
tdst4:
   movb     #4,rotarydwlsel
   movb     #4,rotaryspksel
   bra      tdstdone
tdst8:
   movb     #8,rotarydwlsel
   movb     #8,rotaryspksel

   bra      tdstdone
tdst10:
   movb     #0x10,rotarydwlsel
   movb     #0x10,rotaryspksel

   bra      tdstdone
tdst20:
   movb     #0x20,rotarydwlsel
   movb     #0x20,rotaryspksel

tdstdone:

   ldd      TCNT
   addd     #48 ; slight delay
   std      TC_rotINJ3 ; set up dwell O/C
   bset     TIE,#TFLG_rotINJ3
   movb     #TFLG_rotINJ3, TFLG1

   ldaa     ramdata.testdwell
   ldab     #100
   mul
   addd     #48
   addd     TCNT
   std      TC_rotINJ4  ; set up spark O/C
   bset     TIE,#TFLG_rotINJ4
   movb     #TFLG_rotINJ4, TFLG1
   bra      ir_notest

jLM25:
    jmp     LM25

tst_inj:
   bclr     outpc.squirt, #0x03

; injection counter
   ldx      testmode_cnt
   stx      outpc.istatus5
   beq      LM25
   dex
   stx      testmode_cnt

   ldd     ramdata.testpw ; double check PW != 0
   beq      LM25

tst_inj1:
   ldaa    ramdata.testsel
   anda    #0x0f
   beq     tst_inj1a
   brset   ramdata.testop, #0x60,tst_inj1a  ; all

   bra     tst_inj2

tst_inj1a:
.ifndef MICROSQUIRT
   brclr    seq_inj_ctrl,#1,nopwm1 ; skip if using additional injectors
   movb     InjPWMTim1,pwm1_on
   movb     InjPWMPd1, PWMPER2
tid1:
   movb     PWMPER2, PWMDTY2
   clr      PWMCNT2
   bset     PWME, #0x04 ; enable PWM2
.endif
nopwm1:
   ldx      ramdata.testpw
   beq      tst_inj2   ; bail out if zero

   bset     TCTL2, #0x04
   bset     CFORC, #0x02  ; turn inj on

   ldd      TCNT
   addd     ramdata.testpw
   std      TC1       ; set up injector off event
   bclr     TCTL2, #0x04 ; next mode is off
   bset     TIE, #0x02
   movb     #2, TFLG1
   bset     outpc.squirt, #0x01

tst_inj2:
   ldaa    ramdata.testsel
   anda    #0x0f
   cmpa    #0x01
   beq     tst_inj2a
   brset   ramdata.testop, #0x60,tst_inj2a  ; all

   bra     tst_inj3

tst_inj2a:
.ifndef MICROSQUIRT
   brclr    seq_inj_ctrl,#1,nopwm2 ; skip if using additional injectors   
   movb     InjPWMTim2,pwm2_on
   movb     InjPWMPd2, PWMPER4
   movb     PWMPER4, PWMDTY4  ; PWM enabled but inactive
   clr      PWMCNT4
   bset     PWME, #0x10 ; enable PWM4
.endif
nopwm2:

   bset     TCTL2, #0x40
   bset     CFORC, #0x08  ; turn inj on

   ldd      TCNT
   addd     ramdata.testpw
   std      TC3       ; set up injector off event
   bclr     TCTL2, #0x40 ; next mode is off
   bset     TIE, #0x08
   movb     #0x08, TFLG1
   bset     outpc.squirt, #0x02

tst_inj3:
   ldaa    ramdata.testsel
   anda    #0x0f
   cmpa    #0x02
   beq     tst_inj3a
   brset   ramdata.testop, #0x60,tst_inj3a  ; all

   bra     tst_inj4

tst_inj3a:
.ifndef MICROSQUIRT
   bset     TCTL2, #0x10
.else
   bset     TCTL1, #0x10
.endif
   bset     CFORC, TFLG_rotINJ3  ; turn inj on

   ldd      TCNT
   addd     ramdata.testpw
   std      TC_rotINJ3       ; set up injector off event
.ifndef MICROSQUIRT
   bclr     TCTL2, #0x10 ; next mode is off
.else
   bclr     TCTL1, #0x10 ; next mode is off
.endif
   bset     TIE, #TFLG_rotINJ3
   movb     #TFLG_rotINJ3, TFLG1
   bset     outpc.squirt, #0x04

tst_inj4:
   ldaa    ramdata.testsel
   anda    #0x0f
   cmpa    #0x03
   beq     tst_inj4a
   brset   ramdata.testop, #0x60,tst_inj4a  ; all

   bra     tst_inj_pwm

tst_inj4a:
.ifndef MICROSQUIRT
   bset     TCTL1, #0x01
.else
   bset     TCTL1, #0x40
.endif
   bset     CFORC, TFLG_rotINJ4  ; turn inj on

   ldd      TCNT
   addd     ramdata.testpw
   std      TC_rotINJ4       ; set up injector off event
.ifndef MICROSQUIRT
   bclr     TCTL1, #0x01 ; next mode is off
.else
   bclr     TCTL1, #0x40 ; next mode is off
.endif
   bset     TIE, #TFLG_rotINJ4
   movb     #TFLG_rotINJ4, TFLG1
   bset     outpc.squirt, #0x08

tst_inj_pwm:
.ifndef MICROSQUIRT
; test mode   check for turning on pwm duty cycle
   brclr   seq_inj_ctrl,#1,tst_inj_done ; skip if using additional injectors   
   tst     pwm1_on ; skip if zero
   beq     tsti2

   dec     pwm1_on
   ldaa    pwm1_on
   bne     tsti2

   movb    InjPWMPd1,PWMPER2 ; set PWM period (us)
   movb    pwmd1, PWMDTY2

tsti2:
   tst     pwm2_on ; skip if zero
   beq     tst_inj_done

   dec     pwm2_on
   ldaa    pwm2_on
   bne     tst_inj_done

   movb    InjPWMPd2,PWMPER4 ; set PWM period (us)
   movb    pwmd2, PWMDTY4
.endif
tst_inj_done:
   nop ; space for next mode
   bra      LM25 ; skip over normal inj PWM code

; --------- pulsed i/o test mode --------

tst_pulsedio:
   inc	    mmsDiv_testio
   ldaa	    ramdata.pwm_testio  ; bits 2-3... 78 Hz = 1, 39 Hz = 2, 26 = 3, 19.5 = 4, 15.6 = 5, etc...
   anda	    #0x7
   cmpa	    mmsDiv_testio	
   bhi      end_testio
   clr      mmsDiv_testio
   ldaa	    clk_testio
   inca
   cmpa     #100
   blo      cont_pwm_clktest
   clra
cont_pwm_clktest:
   staa     clk_testio

   ldaa     ramdata.duty_testio
   beq	    testio_off
   cmpa	    clk_testio
   blo      testio_off

testio_on:
    ldx	    port_testio

   ldaa     0,X
   oraa     pin_testio
   staa     0,X

   bra      end_testio

testio_off:
    ldx      port_testio
    ldab     pin_testio
    comb

    andb     0,X 
    stab     0,X

end_testio:

; --------- normal code continues --------

ir_notest:

    jsr     checkfortimerin
    jsr     checkforspk

;    check for turning on pwm duty cycle
.ifndef MICROSQUIRT
   brset   outpc.engine,#2,LM25 ; skip if engine is cranking
   brclr   seq_inj_ctrl,#1,LM25 ; skip if using additional injectors   
   tst     pwm1_on ; skip if zero
   beq     LM20

   dec     pwm1_on
   ldaa    pwm1_on
   bne     LM20

LM17:
   movb    InjPWMPd1, PWMPER2 ; set PWM period (us)
   movb    pwmd1,PWMDTY2

LM20:
   tst     pwm2_on ; skip if zero
   beq     LM25

   dec     pwm2_on
   ldaa    pwm2_on
   bne     LM25

LM22:
   movb    InjPWMPd2, PWMPER4 ; set PWM period (us)
   movb    pwmd2,PWMDTY4
.endif
LM25:
;    // check for re-enabling IC interrupt
;    if(lmms > t_enable_IC)  {
   ldd     lmms
   cpd     t_enable_IC
   bcs     LM7
   bhi     LM6
   ldd     lmms+0x2
   cpd     t_enable_IC+0x2
   bls     LM7

LM6:
   movb    #1, TFLG1   ; clear Ignition IC interrupt flag
   bset    TIE, #1     ; enable Ignition IC interrupt

   movw    #0xffff, t_enable_IC+0x2
   movw    #0xffff, t_enable_IC

LM7:
;    // check for re-enabling IC2 interrupt
;    if(lmms > t_enable_IC2)  {
   ldd     lmms
   cpd     t_enable_IC2
   bcs     LM9
   bhi     LM8
   ldd     lmms+0x2
   cpd     t_enable_IC2+0x2
   bls     LM9

LM8:
   movb    TFLG_trig2, TFLG1   ; clear 2nd trig IC interrupt flag
   bset    TIE, #TFLG_trig2     ; enable 2nd trig IC interrupt

   movw    #0xffff, t_enable_IC2+0x2
   movw    #0xffff, t_enable_IC2

LM9:
;    // idle control
   ldab    IdleCtl
   cmpb    #4
   beq	   pwmidle
   cmpb    #6
   beq     pwmidle
   bra     L31

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

pwmidle:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

new_idle:
    tst     pin_testio
    bne     L31 ; skip in IO testmode
; software pwm idle
   ldaa    flash5.pwmidle_port
   bne     end_pwm_idle

   ldab    idle_dith_cnt
   tfr     b,y

    ldx     pwm_idle_clk
    dex
    stx     pwm_idle_clk
    bne     end_pwm_idle

; now decide whether this is an on or off event
    brclr   pwm_idle_stat, #2, idle_pwm_off

idle_pwm_on:
    ldd     idle_max_on, y
    beq     idle_pwm_off  ; 0% duty
    std     pwm_idle_clk ; next target timer
    bclr    pwm_idle_stat, #2

    ldx	    pPTMpin+0x4 ; pPTMpin[2]
    bset    0,X,#0x04
    bra     end_pwm_idle

idle_pwm_on_sanity:
    ldd     idle_max_on, y
    bne     idle_pwm_on
    bra     end_pwm_idle  ; BOTH are zero...oops

idle_pwm_off:
    ldd     idle_max_off, y
    tfr     d,x
    iny
    iny
    cmpy    #6
    bls     id_o
    ldy     #0
id_o:
    tfr     y,b
    stab    idle_dith_cnt
    tfr     x,d ; grab our saved value back

    cmpd    #0
    beq     idle_pwm_on_sanity  ; 100% duty
    std     pwm_idle_clk ; next target timer
    bset    pwm_idle_stat, #2

    ldx	    pPTMpin+0x4 ; pPTMpin[2]
    bclr    0,X,#0x04

end_pwm_idle:
   movw    IACmotor_pos, outpc.iacstep

;   jmp     L31

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

L31:
   ldaa     conf_err                 ; check for code config error
   beq     LM68  ; skip if no error
   cmpa     #190
   bhs     LM68  ; or >=190  (soft error)

   STACK_SOFT
   call    ign_reset
   UNSTACK_SOFT

   movw    outpc.seconds, outpc.clt ; flag up something weird
   ldd     #60                      ;
   subd    outpc.seconds            ; to catch user's attention
   std     outpc.mat                ;
;   movw    #0xffff, outpc.rpm         ;
   jmp     L47

LM68:

L47:
;    // read fuel sensor data, determine freq(proportional to %alcohol)
   brclr   flash4.FlexFuel,#1,LM95  ; skip if not set
   brset   flash4.FlexFuel,#8,LM95  ; skip if remote

;        if (flash4.flexport == 0) {
   brset   flash4.FlexFuel,#2,ff_pa  ; port a  if 1

;          fsensdat = PORTE & 0x01;   // default  PE0
   ldab    PORTE
   bra     LM79a

ff_pa:
   ldab    PORTA
LM79a:
   andb    #1

;        // other option may be to get GPIO to read frequency of it
   brset   FSensStat,#1,FS1
   brset   FSensStat,#2,FS2

FS0:             ; first time - save pin value (below)
   clr     FPdcounter
   clr     FPdcounter+0x1
   clr     FSens_Pd
   clr     FSens_Pd+0x1
   movb    #1, FSensStat
   bra     LM94

FS1:
;                // wait for pin hi -> lo transition to start period(freq) count
   tst     last_fsensdat ;
   beq     LM94          ;
   tstb                  ;
   bne     LM94          ;

   ldx     FPdcounter    ; .128 ms tics
   inx
   stx     FPdcounter
   movb    #2, FSensStat
   bra     LM94

FS2:
;                // continue period count
   tst     last_fsensdat  ; transition hi -> lo
   beq     LM93           ;
   tstb                   ;
   bne     LM93           ;

   movw    FPdcounter, FSens_Pd      ; save present count = period
   clr     FPdcounter                ; start new period count
   clr     FPdcounter+0x1

LM93:
   ldx    FPdcounter
   inx
   stx    FPdcounter

LM94:
   stab    last_fsensdat    ;  last_fsensdat = fsensdat

;end of FF

LM95:
;2nd trigger polling removed as now input capture

   brset    flagbyte1,#2,skip_dwells
   bra  dwell_timers ; over dwell is always on

skip_dwells:
   jmp      LM112i   ; don't do dwell timers if in test mode

;dwell "actual" timers
;if zero then counter is skipped
;if>0 and <ff then increments
;code can then determine how long a coil has actually charged for
; compared to how long it thought it was being charged for
;In here also see if maxdwl exceeded

dwell_timers:
    ldy     #dwl
    brclr   1,y+, #0xff, LM112b
   dec     -1,y
   bne     LM112b
   ldab    coilsel
   pshb
   movb    #COILABIT,coilsel
   jsr     fire_coil
   pulb
   stab    coilsel
LM112b:
    brclr   1,y+, #0xff, LM112c
   dec     -1,y
   bne     LM112c
   ldab    coilsel
   pshb
   movb    #COILBBIT,coilsel
   jsr     fire_coil  
   pulb
   stab    coilsel
LM112c:
    brclr   1,y+, #0xff, LM112d
   dec     -1,y
   bne     LM112d
   ldab    coilsel
   pshb
   movb    #COILCBIT,coilsel
   jsr     fire_coil  
   pulb
   stab    coilsel
LM112d:
    brclr   1,y+, #0xff, LM112e
   dec     -1,y
   bne     LM112e
   ldab    coilsel
   pshb
   movb    #COILDBIT,coilsel
   jsr     fire_coil  
   pulb
   stab    coilsel   
LM112e:
    brclr   1,y+, #0xff, LM112f
   dec     -1,y
   bne     LM112f
   ldab    coilsel
   pshb
   movb    #COILEBIT,coilsel
   jsr     fire_coil  
   pulb
   stab    coilsel
LM112f:
    brclr   1,y+, #0xff, LM112g
   dec     -1,y
   bne     LM112g
   ldab    coilsel
   pshb
   movb    #COILFBIT,coilsel
   jsr     fire_coil  
   pulb
   stab    coilsel
LM112g:

    ldab    spkmode
    cmpb    #14
    beq     tt_dwell_timer

rdwell_timers:
    ldy     #rdwl
    brclr   1,y+, #0xff, LMrdb
   dec     -1,y
   bne     LMrdb
   ldab    rotaryspksel
   pshb
   movb    #4,rotaryspksel
   jsr     fire_coil_rotary
   pulb
   stab    rotaryspksel
LMrdb:
    brclr   1,y+, #0xff, LMrdc
   dec     -1,y
   bne     LMrdc
   ldab    rotaryspksel
   pshb
   movb    #8,rotaryspksel
   jsr     fire_coil_rotary
   pulb
   stab    rotaryspksel
LMrdc:
    bra LM112i

;; twin trigger overdwell timer
tt_dwell_timer:
    ldy     #rdwl
    brclr   0,y, #0xff, LM112i
   dec     0,y
   bne     LM112i
   jsr     tt2s_sub

LM112i:

   ldd     mapsample_time
   addd    #1
   std     mapsample_time
   ldd     tpssample_time
   addd    #1
   std     tpssample_time

;;;; boost
check_boost:
    tst     pin_testio
    bne     check_countdownmode ; skip in IO testmode
   brclr    flash5.boost_ctl_settings,#0x8,check_countdownmode
   brset    flash5.boost_ctl_settings,#0x40,check_countdownmode  ; Remote boost control
   ldaa	    mmsDiv
   inca 
   staa	    mmsDiv
   ldaa	    flash5.boost_ctl_settings  ; bits 0-2... 78 Hz = 1, 39 Hz = 2, 26 = 3, 19.5 = 4, 15.6 = 5, etc...
   anda	    #0x7
   cmpa	    mmsDiv	
   bhi      check_countdownmode
   movb	    #0,mmsDiv
   ldaa	    boost_ctl_clock
   inca
   cmpa     #100
   blo      cont_boost_clk
   clra
cont_boost_clk:
   staa     boost_ctl_clock

   ldaa     boost_ctl_duty
   beq	    boostOff
   cmpa	    boost_ctl_clock
   blo      boostOff

boostOn:
   brset    flash5.boost_ctl_settings,#0x20,bcClrout
   bra      bcSetout

boostOff:
   brset    flash5.boost_ctl_settings,#0x20,bcSetout
   bra	    bcClrout

bcSetout:
   ldx	    boostport
   ldaa     0,X
   oraa     boostpin
   staa     0,X
   bra      check_countdownmode

bcClrout:
   ldx      boostport
   ldab     boostpin 
   comb
   andb     0,X 
   stab     0,X 

check_countdownmode:
; removed and replaced by dwell queue

    jsr     checkfortimerin
    jsr     checkforspk

dwell_queue:
; array is of ints
; dwellq[0].sel = dwellq
; dwellq[0].time_us  = dwellq+2
; dwellq[0].time_mms = dwellq+4
; dwellq[1].sel = dwellq+6
; dwellq[1].time_us  = dwellq+8
; dwellq[1].time_mms = dwellq+10

dq1:
    ldx     dwellq
    beq     dq2
    ldy     dwellq+4
    dey
    sty     dwellq+4
    bne     dq2
    clr     dwellq
    clr     dwellq+1
    ldd     dwellq+2
    bsr     dq_fire

dq2:
    ldx     dwellq+6
    beq     dq_done
    ldy     dwellq+10
    dey
    sty     dwellq+10
    bne     dq_done
    clr     dwellq+6
    clr     dwellq+7
    ldd     dwellq+8
    bsr     dq_fire

    bra     dq_done

dq_fire:
    xgdx
    stab    dwellsel
    stx     TC_dwl
    movb    #TFLG_dwl, TFLG1
    bset    TIE, #TFLG_dwl
    rts

dq_done:

spark_queue:
; array is of ints
; spkq[0].sel = spkq
; spkq[0].time_us  = spkq+2
; spkq[0].time_mms = spkq+4
; spkq[1].sel = spkq+6
; spkq[1].time_us  = spkq+8
; spkq[1].time_mms = spkq+10

sq1:
    ldx     spkq
    beq     sq2
    ldy     spkq+4
    dey
    sty     spkq+4
    bne     sq2
    clr     spkq
    clr     spkq+1
    ldd     spkq+2
    bsr     sq_fire

sq2:
    ldx     spkq+6
    beq     sq_done
    ldy     spkq+10
    dey
    sty     spkq+10
    bne     sq_done
    clr     spkq+6
    clr     spkq+7
    ldd     spkq+8
    bsr     sq_fire

    bra     sq_done

sq_fire:
    xgdx
    stab    coilsel
    stx     TC_ign
    movb    #TFLG_ign, TFLG1
    bset    TIE, #TFLG_ign
    rts

sq_done:

;;;; start of clocks section

;rtc_clocks:
; are we within 2 rtc ticks of the overflow? if so, let IC know
   ldd TCNT
   cpd #0xFE7F
   blo  really_rtc_clocks
   bset flagbyte1,#flagbyte1_ovfclose


really_rtc_clocks:
;create 0.128ms period timer like MS1 did. Used for tacho out
   ldx     lowres_ctr
   inx
   bne     lr_str
   dex  ; hold at FFFF
lr_str:
   stx     lowres_ctr

;tacho out
   brclr   flash5.tacho_opt,#128,no_tachoff
   cpx     tacho_targ
   bne     no_tachoff
; turn off tacho output
   ldaa    flash5.tacho_opt
   anda    #0x0f
   beq     toff_pt5
   cmpa    #1
   beq     toff_pt7
   cmpa    #2
   beq     toff_pt6
   cmpa    #3
   beq     toff_pa0
   cmpa    #4
   beq     toff_pm2
   cmpa    #5
   beq     toff_pm3
   cmpa    #6
   beq     toff_pm4
   cmpa    #7
   beq     toff_pm5
   cmpa    #8
   beq     toff_ptad40
   cmpa    #9
   beq     toff_ptad80
   bra     no_tachoff ; illegal
toff_pt5:
   bclr    PORTT,#0x20
   bra     no_tachoff
toff_pt6:
   bclr    PORTT,#0x40
   bra     no_tachoff
toff_pt7:
   bclr    PORTT,#0x80
   bra     no_tachoff
toff_pa0:
   bclr    PORTA,#0x01
   bra     no_tachoff
toff_pm2:
   bclr    PORTM,#0x04
   bra     no_tachoff
toff_pm3:
   bclr    PORTM,#0x08
   bra     no_tachoff
toff_pm4:
   bclr    PORTM,#0x10
   bra     no_tachoff
toff_pm5:
   bclr    PORTM,#0x20
   bra     no_tachoff
toff_ptad40:
   bclr    PTAD,#0x40
   bra     no_tachoff
toff_ptad80:
   bclr    PTAD,#0x80
;  bra     no_tachoff

no_tachoff:

; semi-seq time mask

    ldd       inj1cntdown
    beq	      checkinj2
    subd      #1
    std	      inj1cntdown
    beq	      SEMI_SCHED1

checkinj2:
    ldd	      inj2cntdown
    beq	      done_semi
    subd      #1
    std       inj2cntdown
    beq	      SEMI_SCHED2
    bra	      done_semi

SEMI_SCHED1:
    STACK_SOFT
    call      INJ1
    UNSTACK_SOFT
    bra       checkinj2

SEMI_SCHED2:
    STACK_SOFT
    call      INJ2
    UNSTACK_SOFT

done_semi:

; stepper idle PWMed select signal (incl on/off)
    ldaa    pin_iacen
    beq     done_iacen

    ldx     port_iacen
    ldab    iac_dty
    beq     iacenzero
    cmpb    #2
    beq     iacenone
;50%
    eora    0,x
    bra     iacencom

iacenone:
    oraa    0,x
    bra     iacencom
iacenzero:
    coma
    anda    0,x
iacencom:
    staa    0,x
done_iacen:

;    // check mms to generate other clocks
   ldaa	   mms
   cmpa    #7   ; reset every 8 ticks
   bls     CLK_DONE
   clr     mms

; -------------------- 1ms code section --------------------
;milliseconds    (actually 1.024 ms)
   inc     millisec

;    // idle control
   ldab    IdleCtl
   cmpb    #4
   beq	   done_stepper
   cmpb    #6
   beq     done_stepper

   ldab    IAC_moving
   beq     done_stepper

; Time to step?
   tst     motor_time_ms
   beq     iac_move     ; Already zero
   dec     motor_time_ms
   bne     done_stepper      ; Not zero yet.

iac_move:
   cmpb    #2
   beq     iacm2    ; do a step

; IAC_moving = 1
   clr     IAC_moving
    bra     done_stepper

iacm2:
; IAC_moving = 2
   movb    flash4.IACtstep, motor_time_ms

LM45:
   ldd     IACmotor_pos
   cpd     outpc.iacstep
   bgt     step_anticlockwise
   blt     step_clockwise
   bra     LM50

step_clockwise:
   dec     motor_step
   ldab    motor_step
   clra
   andb    #0x7
   stab    motor_step
   ldy     outpc.iacstep
   dey
   sty     outpc.iacstep

   tfr     D,Y

    ldaa    pin_iac1
    ldx     port_iac1
   ldab    IACCoilA,Y
   cmpb     pol_iac1
   beq     LM47a
    oraa    0,x
   bra     LM47b
LM47a:
    coma
    anda    0,x
LM47b:
    staa    0,x

    ldaa    pin_iac2
    ldx     port_iac2
   ldab    IACCoilB,Y
   cmpb     pol_iac2
   beq     LM47c
    oraa    0,x
   bra     LM50
LM47c:
    coma
    anda    0,x
   bra     LM50

step_anticlockwise:
   inc     motor_step
   ldab    motor_step
   clra
   andb    #0x7
   stab    motor_step
   ldy     outpc.iacstep
   iny
   sty     outpc.iacstep

   tfr     D,Y

    ldaa    pin_iac1
    ldx     port_iac1
   ldab    IACCoilA,Y
   cmpb     pol_iac1
   beq     LM49a
    oraa    0,x
   bra     LM49b
LM49a:
    coma
    anda    0,x
LM49b:
    staa    0,x

    ldaa    pin_iac2
    ldx     port_iac2
   ldab    IACCoilB,Y
   cmpb     pol_iac2
   beq     LM49c
    oraa    0,x
   bra     LM50
LM49c:
    coma
    anda    0,x
LM50:
    staa    0,x

   ldd     outpc.iacstep
   cpd     IACmotor_pos
   bne     done_stepper ; skip if not equal

   movb	#1, IAC_moving  ; hold on power for one step-time

LM58:
    movb    iac_holddty, iac_dty ; 0,1,2 = CPU pin is off, 50%, on

done_stepper:
   ldx     ac_idleup_timer
   bmi     b_timer
   inx
   stx     ac_idleup_timer

b_timer:
   brclr   flash5.boost_ctl_settings,#0x8,idle_ctl_timer
   ldd     boost_ctl_timer
   addd    #1
   cmpd    flash5.boost_ctl_ms
   blo     continue_boost
   bset    flagbyte3,#flagbyte3_runboost
   ldd     #0

continue_boost:
   std     boost_ctl_timer

idle_ctl_timer:
   ldab    IdleCtl
   cmpb    #6
   beq     continue_idle_ctl_timer
   cmpb    #7
   beq     continue_idle_ctl_timer
   cmpb    #8
   beq     continue_idle_ctl_timer
   bra     continue_millisec

continue_idle_ctl_timer:
   ldd     pwmidle_timer
   addd    #1
   cmpd    flash5.pwmidle_ms
   blo     store_pwmidle_timer
   bset    flagbyte2,#flagbyte2_runidle
   bset    flagbyte15,#FLAGBYTE15_DPCNT
   ldd	   #0

store_pwmidle_timer:
   std     pwmidle_timer
   
continue_millisec:
   ldab    adc_ctr
   addb    #1
   cmpb    #10   ; every 10.24ms

;50ms now made into 10ms as we were 6x slower than MS1 (8ms sampling period on MAP)
   bne     milli_cont
   clrb

    tst     shift_cut_timer
    beq     no_shift_timer
    dec     shift_cut_timer
no_shift_timer:

   bset    flagbyte20,#FLAGBYTE20_50MS  ; set flag so mainloop can do lag factors

;run nitrous down timer here at 1/100s
   ldaa    n2o_act_timer
   beq     n2chk2
   dec     n2o_act_timer

n2chk2:
   ldaa    n2o2_act_timer
   beq     milli_cont
   dec     n2o2_act_timer

milli_cont:
   stab    adc_ctr
   ldaa    millisec
   cmpa    #98
   bls     CLK_DONE

LM116:
   cmpa    #99
   bls     CLK_DONE
   clr     millisec

;(1/10) tenth of second section
   inc     tpsaclk  ; .10035 sec clocks
   inc     knk_clk
   ldaa    fc_counter
   beq     no_fc
   dec     fc_counter
no_fc:

; reset holdoff timer in 0.1 second units
   tst     resetholdoff
   beq     hod
   dec     resetholdoff
hod:

CLK_DONE:
   ldx    TC_ovflow  ;    // Get display seconds from continuously running TCNT
   inx
   stx    TC_ovflow
   cpx    #7813
   blo    DONE1s
   clr    TC_ovflow
   clr    TC_ovflow+1
   ldx    outpc.seconds ;        // update seconds to send back to PC
   inx
   stx    outpc.seconds
   bclr   outpc.status3, #4 ; STATUS3_MAPERROR reset every 1 second
   bset   flagbyte15,#FLAGBYTE15_ONESEC
   ldaa   bl_timer
   beq    pwmidle_wait_timer
   cmpa   #255
   beq    pwmidle_wait_timer
   inca
   staa   bl_timer

pwmidle_wait_timer:
   ldaa   idle_wait_timer
   cmpa   #255
   beq    idle_advance_tmer 
   inca
   staa   idle_wait_timer

idle_advance_tmer:
   ldaa   idle_advance_timer
   cmpa   #255
   beq    idle_shift_timer 
   inca
   staa   idle_advance_timer

idle_shift_timer:
   ldaa   pwmidle_shift_timer
   cmpa   #255
   beq    running_timer
   inca
   staa   pwmidle_shift_timer

running_timer:
   ldd    outpc.rpm  ; if rpm is 0, don't increment this counter
   beq	  fc_ego_timer 
   ldaa   running_seconds
   cmpa   #255
   beq    fc_ego_timer
   inca
   staa   running_seconds

fc_ego_timer:
   ldaa   fc_off_time
   cmpa   flash5.fc_ego_delay
   bhs    ac_last_on_timer
   inca
   staa   fc_off_time

ac_last_on_timer:
   ldaa   ac_time_since_last_on
   cmpa   #255 
   beq    srl_err_counter
   inca
   staa   ac_time_since_last_on

srl_err_counter:
   tst    srl_err_cnt
   beq    DONE1s
   dec    srl_err_cnt

DONE1s:

   rti

