;*********************************************************************
; Misc asm
;*********************************************************************
; $Id: ms2_extra_asm.s,v 1.21 2013/11/21 13:58:25 jsmcortina Exp $

; * Copyright 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
; * James Murray and Kenneth Culver
; *
; * This file is a part of MS2/Extra.
; *
; * Origin: James Murray
; * Majority: James Murray
; *
; * You should have received a copy of the code LICENSE along with this source,
; * ask on the www.msextra.com forum if you did not.
; *

.sect .text
.globl UnimplementedISR, muldiv, g_read_copy
.globl do_maplog1, do_maplog2, do_maplog_rpm, do_maplog_tooth
.globl do_englog

             nolist               ;turn off listing
             include "ms2extrah.inc"
             list                 ;turn listing back on

;**************************************************************************
; Miscellaneous code in asm
; **************************************************************************
UnimplementedISR:
   ; Unimplemented ISRs trap
   rti

;**************************************************************************
.sect .text
;U32 x U16 / 65536
; put this in low mem, so jsr gets us here
; on entry:
; D is U16
; U32 on stack at 2,SP and 4,SP
; returns result in X:D
muldiv:
    pshd
    clr 1,-sp ; lower word of U32 result 0,SP
    clr 1,-sp ; (no clrw on S12)
;now U16 is 2,SP
;    U32 is 6,SP  8,SP

    ldd     8,SP
    ldy     2,SP
    emul
    sty     0,SP
    ldd     6,SP
    ldy     2,SP
    emul
    addd    0,SP
    std     0,SP
    bcc     noaddy
    iny
noaddy:
    tfr     y,x
    leas    4,SP
    rts

;**************************************************************************

g_read_copy:
; (src, count, dest)
; copies from flash memory to local memory
; on entry D is source address
; first word off stack is count in bytes
; second word off stack is dest address
; returns 1 if changed
; sets/clears interrupt flag
    tfr     d,x  ; source
    ldy     4,sp ; dest
    clr     5,sp ; stat
grc_lp:
    ldd     2,sp ; count
    cmpd    #4
    blo     grc2

;4 bytes
    subd    #4
    std     2,sp

    ldd     2,x+
    cmpd    0,y
    beq     grc1k
    movb    #1,5,sp
grc1k:
    sei
    std     2,y+

    ldd     2,x+
    cmpd    0,y
    beq     grc1k2
    movb    #1,5,sp
grc1k2:
    std     2,y+
    cli
    bra     grc_lp

grc2:
    cmpd    #2
    blo     grc3

;2 bytes
    subd    #2
    std     2,sp

    ldd     2,x+
    cmpd    0,y
    beq     grc2k
    movb    #1,5,sp
grc2k:
    std     2,y+
    bra     grc_lp

grc3:
    cmpd    #1
    blo     grc4
;1 byte
    subd    #1
    std     2,sp

    ldaa    1,x+
    cmpa    0,y
    beq     grc3k
    movb    #1,5,sp
grc3k:
    staa    1,y+
    bra     grc_lp

grc4:
    ldab    5,sp
    rts

; --------------------------------
; map logging subroutine
; called from isr_rtc and from ign_in
; --------------------------------
.sect .text
do_maplog1:
    pshy
    pshx
    pshd
;    bra     do_map_log_com_map

;do_maplog2:
;    pshy
;    pshx
;    pshd
;    xgdx
;    oraa    #0x04
;    xgdx
;;    orx     #0x0400 ; map sensor 2 
;;    bra     do_map_log_com_map

do_map_log_com_map:
; at low rpms, can end up logging too often
    dec     maplog_cnt
    bne     mapldone    ; not yet
    movb    maplog_max, maplog_cnt

do_map_log_com:
    brclr   flagbyte15, FLAGBYTE15_MAPWINDOW, do_map_log_com2
    xgdx
    oraa    #0x40
    xgdx
;    orx     #0x4000
do_map_log_com2:
    ;on entry X contains MAP value to log
    ; bit 15 = 0 means MAP data
    ; MAP sensor no. in bits 12,11,10
    ; bits 9-0 are raw MAP ADC

    ldy	    #ram_data
;S12X    addy	log_offset
    xgdy
    addd    log_offset
    xgdy

    stx	    0,y

    ldy     log_offset
    iny
    iny
    sty	    log_offset
    cpy	    #0x3fc   ; three less
    bls	    mapldone

maplfin:
    ; we've reached the end of the buffer, stop
    bclr	flagbyte0, #FLAGBYTE0_MAPLOG
    bset    outpc.status3, #STATUS3_DONELOG

mapldone:
    puld
    pulx
    puly

    rts
;----

do_maplog_rpm:
    pshy
    pshx
    pshd

    ldx     outpc.rpm
    cpx     #0x3fff
    bls     dmlrk
    ldx     #0x3fff  ; rail
dmlrk:
    xgdx
    oraa    #0xc0
    xgdx
;S12X    orx     #0xc000 ; is rpm
    bra     do_map_log_com2
;----

do_maplog_tooth:
    pshy
    pshx
    pshd
    ldaa    #0x80
    ldab    tooth_no ; byte
    tfr     d,x
;S12X    orx     #0x8000 ; is tooth no. (mainloop with convert to angle at end)
    bra     do_map_log_com2

; --------------------------------
; engine logging subroutine
; called from isr_rtc
; --------------------------------
.sect .text
do_englog:
; at low rpms, can end up logging too often
    dec     maplog_cnt
    bne     engldone    ; not yet
    movb    maplog_max, maplog_cnt

    ldy	    #ram_data
;S12X    addy	log_offset
    xgdy
    addd    log_offset
    xgdy

    ldab    page
    cmpb    #0xf7
    beq     engl_map
    cmpb    #0xf8
    beq     engl_maf

;no MAP or MAF, 2 bytes   
    movb	PTT, 1,y+
    ldab    PTM
    andb    #0x38
    pshb ; temp byte on stack
    ldab    PTAD
    andb    #0xc0
    orab    0,sp
    stab    0,sp
    ldab    PORTA
    andb    #0x01
    orab    0,sp
    stab    0,y ; store
    pulb        ; free up stack

    ldy     log_offset
    bra     englcom ; skips one iny

;MAP or MAF, 3 bytes
engl_map:
    ldd     ATD0DR0
    bra     engl_mapmaf

engl_maf:
    ldx     mafport
    ldd     0,x

engl_mapmaf:
    lsra
    rorb
    lsra
    rorb
    stab    1,y+    ; MAPADC/4     
    movb	PTT, 1,y+
    ldab    PTM
    andb    #0x38
    pshb ; temp byte on stack
    ldab    PTAD
    andb    #0xc0
    orab    0,sp
    stab    0,sp
    ldab    PORTA
    andb    #0x01
    orab    0,sp
    stab    0,y ; store
    pulb        ; free up stack

    ldy     log_offset
    iny
englcom:
    iny
    iny

    sty	    log_offset
    cpy	    #0x3fc   ; three less
    bls	    engldone

englfin:
    ; we've reached the end of the buffer, stop
    bclr	flagbyte0, #FLAGBYTE0_ENGLOG
    bset    outpc.status3, #STATUS3_DONELOG

engldone:
    rts
;----
