/* $Id: ms2_extra_can.c,v 1.37 2014/01/01 00:29:32 racingmini Exp $
 * Copyright 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
 * James Murray and Kenneth Culver
 *
 * This file is a part of MS2/Extra.
 *
 * CanInit()
    Origin: Al Grippo.
    Majority: Al Grippo
 * can_incring()
    Origin: Al Grippo.
    Minor: own function James Murray
    Majority: Al Grippo
 * can_sendburn()
    Origin: Al Grippo.
    Majority: Al Grippo
 * can_reqdata()
    Origin: Al Grippo.
    Majority: Al Grippo
 * can_snddata()
    Origin: Al Grippo.
    Majority: Al Grippo
 * can_xsub01
    Origin: Al Grippo.
    Majority: Al Grippo
 * can_sndMSG_PROT
    Origin: Code by James Murray using MS-CAN framework.
    Majority: James Murray
 * can_sndMSG_SPN
    Origin: Code by James Murray using MS-CAN framework.
    Majority: James Murray
 * can_poll()
    Origin: Jean Belanger (in MS2/Extra)
    Moderate: Additions James Murray
    Majority: Jean Belanger / James Murray
 * can_crc32()
    Origin: Code by James Murray using MS-CAN framework.
    Majority: James Murray
 * ECANSendMessage()
    Origin: "stevevp" used with permission
    Minor: James Murray
    Majority: stevevp
 * can_broadcast()
    Origin: "stevevp" used with permission
    Minor: James Murray
    Majority: stevevp
 *
 * You should have received a copy of the code LICENSE along with this source,
 * ask on the www.msextra.com forum if you did not.
 *
 */
#include "ms2_extra.h"

void CanInit(void)
{
    unsigned int ix;
    /* Set up CAN communications */
    /* Enable CAN, set Init mode so can change registers */
    CANCTL1 |= 0x80;
    CANCTL0 |= 0x01;

    /* clear ring buffers */
    for(ix = 0;ix < 2;ix++)  {
        can[ix].cxno = 0;
        can[ix].cxno_in = 0;
        can[ix].cxno_out = 0;
    }
    can_status = 0;
    flagbyte3 &= ~(flagbyte3_can_reset | flagbyte3_sndcandat | flagbyte3_getcandat);

    while(!(CANCTL1 & 0x01));  // make sure in init mode

    /* Set Can enable, use IPBusclk (24 MHz),clear rest */
    CANCTL1 = 0xC0;  
    /* Set timing for .5Mbits/ sec */
    CANBTR0 = 0xC2;  /* SJW=4,BR Prescaler= 3(24MHz CAN clk) */
    CANBTR1 = 0x1C;  /* Set time quanta: tseg2 =2,tseg1=13 
                        (16 Tq total including sync seg (=1)) */
    CANIDAC = 0x00;   /* 2 32-bit acceptance filters */
    /* CAN message format:
     * Reg Bits: 7 <-------------------- 0
     * IDR0:    |---var_off(11 bits)----|  (Header bits 28 <-- 21)
     * IDR1:    |cont'd 1 1 --msg type--|  (Header bits 20 <-- 15)
     * IDR2:    |---From ID--|--To ID---|  (Header bits 14 <--  7)
     * IDR3:    |--var_blk-|--spare--rtr|  (Header bits  6 <-- 0,rtr)
     */  
    /* Set identifier acceptance and mask registers to accept 
       messages only for can_id or device #15 (=> all devices) */
    /* 1st 32-bit filter bank-to mask filtering, set bit=1 */
    CANIDMR0 = 0xFF;           // anything ok in IDR0(var offset)
    CANIDAR1 = 0x18;           // 0,0,0,SRR=IDE=1
    CANIDMR1 = 0xE7;		   // anything ok for var_off cont'd, msgtype
    CANIDAR2 = CANid;     // rcv msg must be to can_id, but
    CANIDMR2 = 0xF0;			 // can be from any other device
    CANIDMR3 = 0xFF;           // any var_blk, spare, rtr
    /* 2nd 32-bit filter bank */
    CANIDMR4 = 0xFF;           // anything ok in IDR0(var offset)
    CANIDAR5 = 0x18;           // 0,0,0,SRR=IDE=1
    CANIDMR5 = 0xE7;		   // anything ok for var_off cont'd, msgtype
    CANIDAR6 = 0x0F;			 // rcv msg can be to everyone (id=15), and
    CANIDMR6 = 0xF0;			 // can be from any other device
    CANIDMR7 = 0xFF;           // any var_blk, spare, rtr

    /* clear init mode */
    CANCTL0 &= 0xFE;  
    /* wait for synch to bus */
    while ((!(CANCTL0 & 0x10)) && (ix < 0xfffe)) {
        ix++;
        asm("nop");
        asm("nop");
        asm("nop");
        asm("nop");
    }
    if (ix == 0xfffe) { // Did not sync in time
        // CAN broken
        conf_err = 54;
    }

    /* no xmit yet */
    CANTIER = 0x00;
    /* clear RX flag to ready for CAN recv interrupt */
    CANRFLG = 0xC3;
    /* set CAN rcv full interrupt bit */
    CANRIER = 0x01;
    ltch_CAN = 0xffffffff;

    return;
}

unsigned int can_inc_ring0()
{
    unsigned int ret = 0;
    // This is where (in xmt ring buffer) to put next message
    if (can[0].cxno_in < (NO_CANMSG - 1)) {
        can[0].cxno_in++;
    } else {
        can[0].cxno_in = 0;     // overwrite oldest msg in queue
    }
    // increment counter
    if (can[0].cxno < NO_CANMSG) {
        can[0].cxno++;
    } else {
        can[0].cxno = NO_CANMSG;
        ret = 1;
    }
    if (!(CANTIER & 0x07)) {
        // Following will cause entry to TxIsr without sending msg
        // since when CAN0TIER = 0, CAN0TFLG left as buff empty(>0).
        // If CAN0TIER has at least 1 int buf enabled, will enter
        // TxIsr automatically.
        CANTBSEL = CANTFLG;
        CANTIER = CANTBSEL;
    }
    return ret;
}

unsigned int can_sendburn(unsigned int table, unsigned int id)
{
    int ix;
    // set up single CAN message for "burn" & forward to aux board
    // (called from mainloop serial)
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_BURN;
    can[0].cx_destvarblk[ix] = table;
    can[0].cx_destvaroff[ix] = 0;
    can[0].cx_dest[ix] = id;
    can[0].cx_varbyt[ix] = 0;   // no data bytes
    return can_inc_ring0();
}

unsigned int can_reqdata(unsigned int table, unsigned int id, unsigned int offset, unsigned int num)
{
    int ix;
    unsigned int ret;
    // set up single CAN message & forward to aux board
    //   MT requesting data from aux board.
    //   Note: MT must only deal with 1 board at a time and wait til
    //   data back or timeout.
    // (called from mainloop serial)
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_REQ;
    can[0].cx_destvarblk[ix] = table;
    can[0].cx_myvarblk[ix] = 6; // txbuf
    // varblks for the aux boards
    can[0].cx_destvaroff[ix] = offset;
    can[0].cx_myvaroff[ix] = 0;
    can[0].cx_dest[ix] = id;
    can[0].cx_varbyt[ix] = (unsigned char) num;
    ret = can_inc_ring0();
    flagbyte3 |= flagbyte3_getcandat;
    can_getid = id;
    cp_time = (unsigned int)lmms;
    return ret;
}

unsigned int can_snddata(unsigned int table, unsigned int id, unsigned int offset, unsigned int num, unsigned int bufad)
{
    unsigned int ix;
    // set up CAN message & forward to aux board, can be part of chain of packets
    //   PC sending data to aux board
    // called from serial.c and can_isr.c - both use _global_ bufad - requires GPAGE setting externally
    for (ix = 0; ix < num; ix++) {
        can[0].cx_datbuf[can[0].cxno_in][ix] = *(unsigned char*)(bufad + ix);
    }
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_CMD;
    can[0].cx_destvarblk[ix] = table;
    can[0].cx_destvaroff[ix] = offset;
    can[0].cx_dest[ix] = id;
    can[0].cx_varbyt[ix] = num;
    cp_time = (unsigned int)lmms;
    return can_inc_ring0();
}

/* can_t removed */

void can_xsub01(void)
{
    return;
}

unsigned int can_sndMSG_PROT(unsigned int id, unsigned char sz)
{
    int ix;
    unsigned int ret;
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_PROT;
    can[0].cx_destvarblk[ix] = 0;
    can[0].cx_destvaroff[ix] = 0;
    can[0].cx_dest[ix] = id;
    can[0].cx_varbyt[ix] = sz; // either 1 (just protocol no.) or 5 (block sizes as well)
    can[0].cx_myvarblk[ix] = 6; // txbuf - will get sent on down serial line
    can[0].cx_myvaroff[ix] = 0;
    ret = can_inc_ring0();
    flagbyte3 |= flagbyte3_getcandat;
    can_getid = id;
    cp_time = (unsigned int)lmms;
    return ret;
}

unsigned int can_sndMSG_SPND(unsigned int onoff)
{
    int ix;
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_SPND;
    can[0].cx_destvarblk[ix] = 0;
    can[0].cx_destvaroff[ix] = 0;
    can[0].cx_dest[ix] = 15; // broadcast
    can[0].cx_varbyt[ix] = 1;
    can[0].cx_datbuf[ix][0] = onoff;
    return can_inc_ring0();
}

void can_poll_remote(void)
{

    /***************************************************************************
     **
     **  CAN polling from I/O slave device
     **
     **************************************************************************/
    /* CAN message set up as follows
     * cx_msg_type = type of message being sent. MSG_REQ is a request for data
     * cx_dest = remote CANid that we are sending message to
     * cx_destvarblk = which variable block (page) to get data from. page 6 = realtime data, outpc
     * cx_destvaroff = offset within that datablock
     * cx_datbuf = a buffer for the data we are sending, none for a request, eight maximum.
     * cx_varbyt = number of bytes sent or requested
     * cx_myvarblk = the variable block to put the reply in (the other end sends a message back with this block no)
     * cx_myvaroff = the variable offset to put the reply in (the other end sends a message back with this block no)
     *
     * With MSG_REQ we send all of that info off and then the other end replies with a MSG_RSP including the variable block and offset
     * to tell us where to deposit the data, kind of like a "push" of the data. We then (blindly) store it where the MSG_RSP asked us to.
     * In the code below we are asking the GPIO board for the raw ADC data and then store it in outpc.gpioadc[] which are then available for
     * use in the code and are also sent to Megatune as part of the realtime data pack.
     */


    unsigned int ix;
    // load ring buffer - send request for data
    // can[1] is used for mainloop requests
    //CANRIER = 0x00;
//    if ((flash4.can_poll < 3) || ((flash4.can_poll == 3) && (flash4.enable_poll & 0x03))) {
    if ((flash4.can_poll == 1) && (flash4.enable_poll & 0x03)) {
        // Get remote ADC data (GPIO, MS2/Extra slave, Generic board with ADC polling enabled)
        for (ix=0 ; ix<2 ; ix++) { // loop to make two 8 byte transfer for all 8 words
            if (flash4.enable_poll & (ix+1)) {
                can[1].cx_msg_type[can[1].cxno_in] = MSG_REQ;
                if (ix == 0)
                    can[1].cx_dest[can[1].cxno_in] = (flash4.can_poll_id0 & 0x0F);            // send to device 1
                else
                    can[1].cx_dest[can[1].cxno_in] = (flash4.can_poll_id1 & 0x0F);            // send to device 2

//                if (flash4.can_poll == 1) {
//                    can[1].cx_destvarblk[can[1].cxno_in] = 2;   // fetch from block 2 - GPIO
//                    can[1].cx_destvaroff[can[1].cxno_in] = 2 + (ix*8);  // fetch from adcs start at byte 2 - GPIO
//                } else if (flash4.can_poll == 2) {
//                    can[1].cx_destvarblk[can[1].cxno_in] = 7;   // fetch from block 7 - MS2 + MS2/Extra
//                    can[1].cx_destvaroff[can[1].cxno_in] = (unsigned short)(&outpc.gpioadc[0]) - (unsigned short)(&outpc) + (ix*8);  // fetch raw adc from outpc as MS2/Extra slave should have stored them here
//                } else if (flash4.can_poll == 3) {
                    can[1].cx_destvarblk[can[1].cxno_in] = flash4.poll_tables[ix];   // fetch from user defined table
                    can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[ix];  // fetch raw adc from user defined offset
//                }
                can[1].cx_varbyt[can[1].cxno_in] = 8;    // 8 bytes to be returned

                // where should the resulting data be stored
                can[1].cx_myvarblk[can[1].cxno_in] = 7;  // store returned data in outpc
                can[1].cx_myvaroff[can[1].cxno_in] = (unsigned short)(&outpc.gpioadc[0]) - (unsigned short)(&outpc) + (ix*8); // this is offset of where to store it

                // This is where (in xmt ring buffer) to put next message
                if(can[1].cxno_in < (NO_CANMSG - 1)) {
                    can[1].cxno_in++;
                } else {
                    can[1].cxno_in = 0;
                }
                // increment counter
                if(can[1].cxno < NO_CANMSG) {
                    can[1].cxno++;
                } else {
                    can[1].cxno = NO_CANMSG;
                }
            }
        }
    }
    if ((flash4.can_poll == 1) && (flash4.enable_poll & 0x04)) {
        // Get remote PWM data from Generic board
        can[1].cx_msg_type[can[1].cxno_in] = MSG_REQ; 
        can[1].cx_dest[can[1].cxno_in] = (flash4.can_poll_id2 & 0x0F);            // send to device 3

        can[1].cx_destvarblk[can[1].cxno_in] = flash4.poll_tables[2];  // fetch from user defined table
        can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[2];  // fetch raw pwm from user defined offset
        can[1].cx_varbyt[can[1].cxno_in] = 8;    // 8 bytes to be returned

        // where should the resulting data be stored
        can[1].cx_myvarblk[can[1].cxno_in] = 7;  // store returned data in outpc
        can[1].cx_myvaroff[can[1].cxno_in] = (unsigned short)(&outpc.gpiopwmin[0]) - (unsigned short)(&outpc); // this is offset of where to store it

        // This is where (in xmt ring buffer) to put next message
        if(can[1].cxno_in < (NO_CANMSG - 1)) {
            can[1].cxno_in++;
        } else {
            can[1].cxno_in = 0;
        }
        // increment counter
        if(can[1].cxno < NO_CANMSG) {
            can[1].cxno++;
        } else {
            can[1].cxno = NO_CANMSG;
        }
    }
    if ((flash4.can_poll == 1) && (flash4.enable_poll & 0x08) && ((flash4.ports_dir & 0x0F) < 7)) {
        // Get remote ports data from Generic board
        can[1].cx_msg_type[can[1].cxno_in] = MSG_REQ; 
        can[1].cx_dest[can[1].cxno_in] = (flash4.can_poll_id3 & 0x0F);            // send to device 4

        can[1].cx_destvarblk[can[1].cxno_in] = flash4.poll_tables[3];  // fetch from user defined table
        if ((flash4.ports_dir & 0x0F) == 0) {
            // 3 inputs
            can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[3];  // fetch raw port from user defined offset
            can[1].cx_varbyt[can[1].cxno_in] = 3;    // 3 bytes to be returned
            ix = 0;
        } else if ((flash4.ports_dir & 0x0F) == 1) {
            // 2 inputs
            can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[3] + 1;  // fetch raw port from user defined offset
            can[1].cx_varbyt[can[1].cxno_in] = 2;    // 2 bytes to be returned
            ix = 1;
        } else {
            // 1 input
            can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[3] + 2;  // fetch raw port from user defined offset
            can[1].cx_varbyt[can[1].cxno_in] = 1;    // 1 bytes to be returned
            ix = 2;
        }

        // where should the resulting data be stored
        can[1].cx_myvarblk[can[1].cxno_in] = 7;  // store returned data in outpc
        can[1].cx_myvaroff[can[1].cxno_in] = (unsigned short)(&outpc.gpioport[ix]) - (unsigned short)(&outpc); // this is offset of where to store it

        // This is where (in xmt ring buffer) to put next message
        if(can[1].cxno_in < (NO_CANMSG - 1)) {
            can[1].cxno_in++;
        } else {
            can[1].cxno_in = 0;
        }
        // increment counter
        if(can[1].cxno < NO_CANMSG) {
            can[1].cxno++;
        } else {
            can[1].cxno = NO_CANMSG;
        }
    }
    if ((flash4.can_poll == 1) && (flash4.enable_poll & 0x08) && (flash4.ports_dir & 0x0F)) {
        // Update PWM idle if used
        if (flash5.pwmidle_port) {
            outpc.gpioport[flash5.pwmidle_port-1] = (unsigned char)outpc.iacstep;
        }
        // Set remote ports data on Generic board
        can[1].cx_msg_type[can[1].cxno_in] = MSG_CMD; 
        can[1].cx_dest[can[1].cxno_in] = (flash4.can_poll_id3 & 0x0F);            // send to device 4

        can[1].cx_destvarblk[can[1].cxno_in] = flash4.poll_tables[3];  // put to user defined table
        can[1].cx_destvaroff[can[1].cxno_in] = flash4.poll_offset[3];  // put raw port to user defined offset
        if ((flash4.ports_dir & 0x0F) == 1) {
            // 1 output
            can[1].cx_varbyt[can[1].cxno_in] = 1;    // 1 bytes to be returned
        } else if ((flash4.ports_dir & 0x0F) == 3) {
            // 2 outputs
            can[1].cx_varbyt[can[1].cxno_in] = 2;    // 2 bytes to be returned
        } else {
            // 3 outputs
            can[1].cx_varbyt[can[1].cxno_in] = 3;    // 3 bytes to be returned
        }

        for(ix = 0;ix < can[1].cx_varbyt[can[1].cxno_in];ix++)  {
            can[1].cx_datbuf[can[1].cxno_in][ix] = *((unsigned char *)outpc.gpioport + ix);
        }

                // This is where (in xmt ring buffer) to put next message
        if(can[1].cxno_in < (NO_CANMSG - 1)) {
            can[1].cxno_in++;
        } else {
            can[1].cxno_in = 0;
        }
        // increment counter
        if(can[1].cxno < NO_CANMSG) {
            can[1].cxno++;
        } else {
            can[1].cxno = NO_CANMSG;
        }
    }
    if(!(CANTIER & 0x07))  {
        // Following will cause entry to TxIsr without sending msg
        // since when CANTIER = 0, CANTFLG left as buff empty(>0).
        // If CANTIER has at least 1 int buf enabled, will enter
        // TxIsr automatically.
        CANTBSEL = CANTFLG;
        CANTIER = CANTBSEL;
    }
    //CANRIER = 0x01;
}

unsigned int can_crc32(unsigned int table, unsigned int id)
{
    int ix;
    unsigned int ret;
    // set up single CAN message & forward to aux board
    //   PC requesting CRC from aux board.
    // (called from mainloop serial with ints off)
    ix = can[0].cxno_in;
    can[0].cx_msg_type[ix] = MSG_CRC; 
    can[0].cx_destvarblk[ix] = table;
    can[0].cx_myvarblk[ix] = 6;  // txbuf
    // varblks for the aux boards
    can[0].cx_destvaroff[ix] = 0;
    can[0].cx_myvaroff[ix] = 0;
    can[0].cx_dest[ix] = id;
    can[0].cx_varbyt[ix] = 4; // 4 bytes of CRC
    ret = can_inc_ring0();
    flagbyte3 |= flagbyte3_getcandat;
    can_getid = id;
    cp_time = (unsigned int)lmms;
    return ret;
}

/* The following function was written by 'stevevp' and donated to MS2/Extra / MS3 */
void ECANSendMessage(unsigned int id, unsigned char *data, unsigned char dataLen, unsigned char msgType) {
    int ix, kx;
    /* CAN message format:
     * Reg Bits: 7<------------------- 0
     * IDR0:    |----STD ID(11 bits)----|  (Header bits 28 <-- 21)
     * IDR1:    |cont'd 0 0 ------------|  (Header bits 20 <-- 15)
     * IDR2:    |-----------------------|  (Header bits 14 <--  7)
     * IDR3:    |----------------------0|  (Header bits  6 <-- 0,rtr)
     */ 

    ix = can[0].cxno_in;

    // Set to Standard 11Bit ID Format
    can[0].cx_msg_type[ix] = msgType; // MSG_STD = 11

    // set CANID and DLC of Message
    can[0].cx_destvaroff[ix] = id; // CANID
    can[0].cx_varbyt[ix] = dataLen; // 8 bytes DLC

    // load Message Data
                    for (kx = 0;kx < dataLen ;kx++)  {
                        can[0].cx_datbuf[ix][kx] = data[kx];
                    }

    // set following to zero as not used in Standard CAN Header
    can[0].cx_destvarblk[ix] = 0;
    can[0].cx_dest[ix] = 0;

    // This is where (in xmt ring buffer) to put next message
    if (can[0].cxno_in < (NO_CANMSG - 1)) {
        can[0].cxno_in++;
    } else {
        can[0].cxno_in = 0; // overwrite oldest msg in queue
    }

    // increment counter
    if (can[0].cxno < NO_CANMSG) {
        can[0].cxno++;
    } else {
        can[0].cxno = NO_CANMSG;
    }
    if (!(CANTIER & 0x07)) {
        // Following will cause entry to TxIsr without sending msg
        // since when CANTIER = 0, CANTFLG left as buff empty(>0).
        // If CANTIER has at least 1 int buf enabled, will enter
        // TxIsr automatically.
        CANTBSEL = CANTFLG;
        CANTIER = CANTBSEL;
    }
}

/* The following function is derived from code written by 'stevep' and donated to MS2/Extra / MS3 */
void can_broadcast(void)
{
    /* Broadcast selected CAN messages using STD CAN*/

    // Send Message Variables
    unsigned int id, val;
    unsigned char data[8];
    unsigned char dataLen;

    DISABLE_INTERRUPTS;
    val = (unsigned int)lmms;
    ENABLE_INTERRUPTS;
    if ((val - can_bcast_last) > flash4.can_bcast_int) {
        can_bcast_last = val; // rollover safe
    } else {
        return;
    }
    
    if (flash4.can_bcast1 & 0x02) {
        id = Engine_RPM; //0x280 - Motorsteuergerat 640 - engine control unit RPM is 4:1 ratio = outpc.rpm*4
        val = outpc.rpm * 4;
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = 0x00;
        data[2] = (unsigned char)(val & 0xff); //byte 3 = RPM, L
        data[3] = (unsigned char)(val >> 8);   //byte 4 = RPM, H
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if (flash4.can_bcast1 & 0x04) {
        /* id definition says 1:1 ?? */
        id = Engine_RPM; //0x280 - Motorsteuergerat 640 - engine control unit RPM is 1:1 ratio = outpc.rpm*1
        val = outpc.rpm * 1;
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = 0x00;
        data[2] = (unsigned char)(val & 0xff); //byte 3 = RPM, L
        data[3] = (unsigned char)(val >> 8);   //byte 4 = RPM, H
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if (flash4.can_bcast1 & 0x08) {
        long t1, t2;
        id = Engine_Temp;//     0x289 // 996 - Motorsteuerger�t 649 - Engine Temp /y=46.14+1.67*x where x is degrees C
        // quadratic to match guage position curve y = -.0336*x*x + 7.7986*x + -225.2571
        val = (((outpc.clt - 320) * 5) / 9)/10;
        if (val > 74) {
            t1 = (-336L * val) * val;
            t2 = 77986L * val;
            val = (unsigned int)((t1 + t2 -2252571) / 10000);
        }
        else {
            val = (5447 + (154 * val)) / 100;
        }
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = 0x00;
        data[2] = (unsigned char)val;
        data[3] = 0x00;
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if (flash4.can_bcast1 & 0x10) { // from Peter Florance
	    id = Engine_RPM316; //0x316 BMW E46 - engine control unit RPM is 6.42:1 ratio = outpc.rpm*6.42
        val = (unsigned int)((outpc.rpm * 642UL) / 100);        
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = 0x00;
        data[2] = (unsigned char)(val & 0xff); //byte 3 = RPM, L
        data[3] = (unsigned char)(val >> 8);   //byte 4 = RPM, H
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if (flash4.can_bcast1 & 0x20) { // from Peter Florance
        id = Engine_Temp329;//     0x329 
//        val = (((outpc.clt - 320) * 5) / 9)/10;
//	    val = (int)(((val + 48)*4)/3);
        val = (1088 + (outpc.clt << 1)) / 27;        
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = (unsigned char)val;
        data[2] = 0x00;
        data[3] = 0x00;
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if ((flash4.can_bcast1 & 0x40)
        && (((flash4.can_bcast2 & 0x01) == 0) || (flagbyte6 & FLAGBYTE6_CAN1ST))) { // and 1st message sent (if enabled)
        /* Alfa/Fiat/Lancia dash coolant, RPM. Submitted by 'acab' */
        int valclt, valrpm;
        id = 0x561;

        val = (outpc.clt - 320) / 18;
        valclt = val + 40;
        valrpm = outpc.rpm/32;
        
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = 0x00;
        data[2] = 0x00;
        data[3] = (unsigned char)valclt;   // Coolant temp
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = (unsigned char)valrpm;   // RPM 
        data[7] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }
/* additional submissions by 'acab'
    0x361 | 8 | 00 xx 00 00 00 00 00 00
    xx= torque 0-99.96 % / 1bit = 0.392 %

    0x361 | 8 | 00 00 xx xx 00 00 00 00
    xx xx = rpm 0-10240 rpm / 1bit = 1 rpm

    0x361 | 8 | 00 00 00 00 00 00 00 xx
    xx= gas pedal position 0-99,96 % / 1bit = 0,392 % (may be same as throttle position)

    0x361 | 8 | 00 00 00 00 xx 00 00 00
    xx= throttle position 0- 99,96 % / 1bit = 0,392 %

    Initialisation packet, send only one time after ignition:
    0x041 | 7 | 06 00 00 00 00 00 00
*/
    if ((flash4.can_bcast1 & 0x80) 
        && (((flash4.can_bcast2 & 0x01) == 0) || (flagbyte6 & FLAGBYTE6_CAN1ST))) { // and 1st message sent (if enabled)
        int tps, torque;
        id = 0x361;

        if (outpc.tps < 0) {
            tps = 0;
        } else if (outpc.tps > 1000) {
            tps = 255;
        } else {
            tps = (outpc.tps *255UL) / 1000;
        }

        torque = calc_duty(global_base_pw1); // assume duty is a reasonable measure of torque
        dataLen = 0x08;
        data[0] = 0x00;
        data[1] = (unsigned char)torque;
        data[2] = outpc.rpm >> 8;
        data[3] = outpc.rpm & 0xff;
        data[4] = (unsigned char)tps;
        data[5] = 0x00;
        data[6] = 0x00;
        data[7] = (unsigned char)tps;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }

    if ((flash4.can_bcast2 & 0x01) && (outpc.seconds) && (!(flagbyte6 & FLAGBYTE6_CAN1ST))) {
        flagbyte6 |= FLAGBYTE6_CAN1ST;
        id = 0x041;

        dataLen = 0x07;
        data[0] = 0x06;
        data[1] = 0x00;
        data[2] = 0x00;
        data[3] = 0x00;
        data[4] = 0x00;
        data[5] = 0x00;
        data[6] = 0x00;
        ECANSendMessage(id, data, dataLen, MSG_STD);
    }
}
