/* $Id: ms2_extra_can_isr.c,v 1.35 2013/12/31 13:50:34 jsmcortina Exp $
 * Copyright 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
 * James Murray and Kenneth Culver
 *
 * This file is a part of MS2/Extra.
 *
 * CanTxIsr()
    Origin: Al Grippo.
    Moderate: Additions by James Murray / Jean Belanger
    Majority: Al Grippo / James Murray / Jean Belanger
 * CanRxIsr()
    Origin: Al Grippo.
    Moderate: Additions by James Murray / Jean Belanger
    Majority: Al Grippo / James Murray / Jean Belanger
 *
 * You should have received a copy of the code LICENSE along with this source, please
 * ask on the www.msextra.com forum if you did not.
 *
*/

#include "ms2_extra.h"

INTERRUPT void CanTxIsr(void)
{
    unsigned char ix,jx,kx,lx,var_byt,cur_size,msg_no,new_in;

    /* CAN Xmit Interrupt */
    CANTBSEL = CANTFLG;    // select MSCAN xmit buffer
    // Check ring buffers and xfer to MSCAN buffer
    for (ix = 0;ix < 2;ix++)  {
        if (can[ix].cxno)  {
            unsigned char msg_type, destblk;

            jx = can[ix].cxno_out;
            msg_type = can[ix].cx_msg_type[jx];
            destblk = can[ix].cx_destvarblk[jx];

            if (msg_type != MSG_STD) {
                /* Set up identifier registers */
                CAN_TB0_IDR0 = (unsigned char)(can[ix].cx_destvaroff[jx] >> 3);
                // 8 high bits in IDR0, 3 low bits in IDR1
                CAN_TB0_IDR1 = (unsigned char)
                        ((can[ix].cx_destvaroff[jx] & 0x0007) << 5)
                        | 0x18                              // SRR=IDE=1
                        | (can[ix].cx_msg_type[jx] & 0x07);	// 3 bits
                CAN_TB0_IDR2 = (CANid << 4) | can[ix].cx_dest[jx];
                // Dest table with additional bit
                CAN_TB0_IDR3 = (destblk << 4)
                                | ((destblk & 0x10) >> 1);
            }

            /* Set xmt buffer priorities (lower is > priority) */
            CAN_TB0_TBPR = 0x02;

            /* set data in buffer */
            switch (msg_type)  {
                case MSG_CMD:  // msg for dest ecu to set a variable to val in msg
                case MSG_RSP:  // msg in reply to dest ecu's request for a var val
                    CAN_TB0_DLR = can[ix].cx_varbyt[jx];
                    for (kx = 0;kx < CAN_TB0_DLR;kx++)  {
                        *(&CAN_TB0_DSR0 + kx) = can[ix].cx_datbuf[jx][kx];
                    }
                    break;

                case MSG_REQ:  // msg to send back current value of variable(s)
                    // this 1st byte holds var blk for where to put rcvd data
                    // 2nd,3rd bytes hold var offset rel. to var blk and how
                    // many consecutive bytes to be sent back
                    if (destblk <= 31) {
                        CAN_TB0_DLR = 3;
                        CAN_TB0_DSR0 = can[ix].cx_myvarblk[jx];
                        CAN_TB0_DSR1 = (unsigned char) (can[ix].cx_myvaroff[jx] >> 3);
                        CAN_TB0_DSR2 = (unsigned char) ((can[ix].cx_myvaroff[jx] & 0x0007) << 5) | can[ix].cx_varbyt[jx];
                    } else {
                        CAN_TB0_IDR1 = (CAN_TB0_IDR1 & 0xf8) | MSG_XTND; 
                        CAN_TB0_DLR = 5;
                        CAN_TB0_DSR0 = MSG_REQX;
                        CAN_TB0_DSR1 = can[ix].cx_myvarblk[jx];
                        CAN_TB0_DSR2 = (unsigned char)(can[ix].cx_myvaroff[jx] >> 3);
                        CAN_TB0_DSR3 = (unsigned char)((can[ix].cx_myvaroff[jx] & 0x0007) << 5) | can[ix].cx_varbyt[jx];
                        CAN_TB0_DSR4 = destblk;
                    }
                    break;

//                case OUTMSG_REQ:  // Not implemented

                case OUTMSG_RSP:  // msg in reply to dest ecu's request for an outmsg 
                    CAN_TB0_DLR = can[ix].cx_varbyt[jx];
                    for (kx = 0;kx < CAN_TB0_DLR;kx++)  {
                        *(&CAN_TB0_DSR0 + kx) = can[ix].cx_datbuf[jx][kx];
                    }

                    msg_no = can[ix].cx_myvarblk[jx];   // Message number stored in myvarblk

                    // Compute the number of bytes and set the data for the next response packet
                    var_byt = 0;
                    new_in = can[0].cxno_in;
                    for (kx = cur_outmsg_var[msg_no]; kx < OUTMSG_SIZE; ) {
                        cur_size = flash12.outmsgs[msg_no].size[kx];
                        if ((cur_size != 0) && ((var_byt + cur_size) <= 8)) {
                            for (lx = 0; lx < cur_size; lx++) {
                                can[0].cx_datbuf[new_in][var_byt] = *((unsigned char *)&outpc + flash12.outmsgs[msg_no].offset[cur_outmsg_var[msg_no]] + (unsigned int)lx);
                                var_byt++;
                            }
                            cur_outmsg_var[msg_no]++;
                        } else {
                            break;
                        }
                        kx += cur_size;
                    }

                    if (var_byt) {
                        // Set the next message
                        unsigned char dest, var_blk;
                        unsigned short var_off;

                        dest = can[ix].cx_dest[jx];
                        var_blk = can[ix].cx_destvarblk[jx]; 
                        var_off = can[ix].cx_destvaroff[jx]; 
                        can[0].cx_msg_type[new_in] = OUTMSG_RSP;
                        can[0].cx_myvarblk[new_in] = msg_no;     // Use the local var block as the message number
                        can[0].cx_destvarblk[new_in] = var_blk;  // Use the same destination var_blk
                        can[0].cx_destvaroff[new_in] = var_off + CAN_TB0_DLR; // Use the current var_off plus the current message size as the new destination offset
                        can[0].cx_dest[new_in] = dest;           // Keep the same desitnation
                        can[0].cx_varbyt[new_in] = var_byt;

                        // This is where (in xmt ring buffer) to put next message
                        if (can[0].cxno_in < (NO_CANMSG - 1)) {
                            can[0].cxno_in++;
                        } else {
                            can[0].cxno_in = 0;     // overwrite oldest msg in queue
                        }
                        // increment counter (which will be decremented at the end of the ISR so no boundary check is needed)
                        can[0].cxno++;
                    }
                    break;
		  
                case MSG_XSUB:
                    CAN_TB0_DLR = 0;
                    break;

                case MSG_BURN:
                    CAN_TB0_DLR = 0;
                    break;

                case MSG_CRC:
                    // convert to extended format
                    CAN_TB0_IDR1 = (CAN_TB0_IDR1 & 0xf8) | MSG_XTND; 
                    CAN_TB0_DLR = 4;
                    CAN_TB0_DSR0 = MSG_CRC;
                    CAN_TB0_DSR1 = can[ix].cx_myvarblk[jx];
                    CAN_TB0_DSR2 = (unsigned char)(can[ix].cx_myvaroff[jx] >> 3);
                    CAN_TB0_DSR3 = (unsigned char)((can[ix].cx_myvaroff[jx] & 0x0007) << 5) | can[ix].cx_varbyt[jx];
                    break;

                case MSG_PROT:
                    // convert to extended format
                    CAN_TB0_IDR1 = (CAN_TB0_IDR1 & 0xf8) | MSG_XTND; 
                    CAN_TB0_DLR = 4;
                    CAN_TB0_DSR0 = MSG_PROT;
                    CAN_TB0_DSR1 = can[ix].cx_myvarblk[jx];
                    CAN_TB0_DSR2 = (unsigned char)(can[ix].cx_myvaroff[jx] >> 3);
                    CAN_TB0_DSR3 = (unsigned char)((can[ix].cx_myvaroff[jx] & 0x0007) << 5) | can[ix].cx_varbyt[jx];
                    break;

                case MSG_SPND:
                    // convert to extended format
                    CAN_TB0_IDR1 = (CAN_TB0_IDR1 & 0xf8) | MSG_XTND; 
                    CAN_TB0_DLR = 2;
                    CAN_TB0_DSR0 = MSG_SPND;
                    CAN_TB0_DSR1 = can[ix].cx_datbuf[jx][0];
                    break;

                case MSG_STD:  // msg format 11 bit STD - do not let MSII mess with it
                    // reset identifier registers for standard ID
                    CAN_TB0_IDR0 = (unsigned char)(can[ix].cx_destvaroff[jx] >> 3);
                    // 8 high bits in IDR0, 3 low bits in IDR1
                    CAN_TB0_IDR1 = (unsigned char)((can[ix].cx_destvaroff[jx] & 0x0007) << 5);
                    CAN_TB0_IDR2 = (unsigned char)(0x00);
                    CAN_TB0_IDR3 = (unsigned char)(0x01);
                    CAN_TB0_DLR = (unsigned char) (can[ix].cx_varbyt[jx]);
                    for (kx = 0;kx < CAN_TB0_DLR;kx++)  {
                        *(&CAN_TB0_DSR0 + kx) = can[ix].cx_datbuf[jx][kx];
                    }
                    break;
            }
            // This is where (in xmt buffer) to get next outgoing message
            if (can[ix].cxno_out < (NO_CANMSG - 1)) {
                can[ix].cxno_out++;
            } else {
                can[ix].cxno_out = 0;
            }
            if (can[ix].cxno > 0) {
                can[ix].cxno--;    // TB buf loaded for xmit - decrement ring buf count
            }
            can_status &= CLR_XMT_ERR;
            /* set CAN xmt interrupt bit and clear flag to initiate transmit */
            CANTIER |= CANTBSEL;  
            CANTFLG = CANTBSEL;  // 1 clears(buffer full), 0 is ignored
            break;               // only handle 1 buffer entry at time
        }
        if (ix == 1)  {    // nothing left in either ring buffer
            CANTIER = 0x00;  // leave CANTFLG as buffer empty, but disable interrupt
            // Note: if this last xmt in buf, will re-neter ISR, but 
            //  then exit because can[0,1].cxno = 0. 
            break;  
        }
    }		        // end for loop

    /* check for any passthrough write messages */
    if ((flagbyte3 & flagbyte3_sndcandat) && cp_targ) {
        int num;
        unsigned int r;

        num = cp_targ - cp_cnt;
        if (num > 8) {
            num = 8;
        } else {
            /* final block, prevent sending any more*/
            flagbyte14 |= FLAGBYTE14_SERIAL_OK; // serial code to return ack code that all bytes were stuck into pipe
            flagbyte3 &= ~flagbyte3_sndcandat;
            cp_targ = 0;
        }
        
        r = can_snddata(cp_table, cp_id, cp_offset + cp_cnt, num, (unsigned int)&srlbuf[9 + cp_cnt]); // ring0
        if (r) {
            /* flag up the error so serial.c can act*/
            flagbyte14 |= FLAGBYTE14_CP_ERR;
        }
        cp_cnt += num;
    }

    /* check for any passthrough read messages */
    if ((flagbyte3 & flagbyte3_getcandat) && cp_targ) {
        int num;

        num = cp_targ - cp_cnt;
        if (num > 8) {
            num = 8;
        } else {
            /* final block, prevent sending any more*/
            cp_targ = 0;
        }

        if (can_reqdata(cp_table, cp_id, cp_offset + cp_cnt, num)) {
            /* flag up the error so serial.c can act*/
            flagbyte14 |= FLAGBYTE14_CP_ERR;
        }
        cp_cnt += num;
        cp_time = (unsigned int)lmms;
    }

    if ((CANRFLG & 0x0C) != 0)  {
        // Xmt error count
        can_status |= XMT_ERR;
        //can_reset = 1;
    }
    return;
}

INTERRUPT void CanRxIsr(void)
{
    unsigned char rcv_id,msg_type,var_blk,var_byt,jx,kx,lx,cur_size,msg_no;
    unsigned short var_off,dvar_off;

    /* CAN Recv Interrupt */
    if (CANRFLG & 0x01)  {
        var_off = ((unsigned short)CAN_RB_IDR0 << 3) |
            ((CAN_RB_IDR1 & 0xE0) >> 5);
        msg_type = (CAN_RB_IDR1 & 0x07);
        if(msg_type == MSG_XTND)
            msg_type = CAN_RB_DSR0;
        rcv_id = CAN_RB_IDR2 >> 4;		// message from device rcv_id
        var_blk = (CAN_RB_IDR3 >> 4) | ((CAN_RB_IDR3 & 0x08) << 1);
        var_byt = (CAN_RB_DLR & 0x0F);

        if (var_off >= 0x4000) {
            /* Impossible! */
            /* in ver 2 protocol will send back error reply */
            goto CAN_DONERXSW;
        }

        switch(msg_type)  {
            case MSG_CMD:  // msg for this ecu to set a variable to val in msg
            case MSG_RSP:  // msg in reply to this ecu's request for a var val
                // value in the data buffer in recvd msg
            if ((msg_type == MSG_RSP) && (flagbyte3 & flagbyte3_getcandat) && (var_blk == 6))  {
                // only forward on via serial if doing CAN stuff and requested to dump into txbuf
                // set up for serial xmit of the recvd CAN bytes (<= 8)
                //  MT getting back data requested from aux board

                if (var_byt && (var_byt <= 8)) {
                    for (jx = 0; jx < var_byt; jx++) {
                        srlbuf[canrxad++] = *(&CAN_RB_DSR0 + jx);
                        if (canrxgoal) {
                            canrxgoal--;
                        }
                    }
                    if (canrxgoal == 0) {
                        /* not expecting any more data */
                        flagbyte15 |= FLAGBYTE15_CANRX; /* tell mainloop to do something with the data */
                    }
                } else {
                    /* how can we get here ? */
                    /* turn off CAN receive mode */
                    flagbyte3 &= ~flagbyte3_getcandat;
                }

                ltch_CAN = 0xFFFFFFFF;

                } else  {
                    unsigned int dest_addr = 0, ad_tmp = 0;
                    unsigned char ck = 0;

                    if (var_blk <= 3) {
                        if ((pg8_ptr->flashlock & 1) == 0) {
                            dest_addr = 0;
                            /* ideally want to send an error code back to the sender */
                        } else {
                            chknewpage(var_blk);
                            dest_addr = (unsigned int)&ram_data;
                            ad_tmp = dest_addr + (var_off & 0x3ff);
                        }
                    } else if (var_blk == 6) {
                        dest_addr = (unsigned int)&canbuf;
                        ad_tmp = dest_addr + var_off;

                    } else if (var_blk == 7) {
                        // special case for receiving CAN data
                        if (var_off >= 0x200) { // datax1 sharing same block no. as outpc
                            var_off -= 0x200; // but at addresses 0x200 and beyond
                            if ((var_off + var_byt) > sizeof(datax1)) {
                                flagbyte3 |= flagbyte3_can_reset;
                            } else {
                                dest_addr = (unsigned int) &datax1;
                            }
                        } else {
                            if ((var_off + var_byt) > sizeof(outpc)) {
                                flagbyte3 |= flagbyte3_can_reset;
                            } else {
                                dest_addr = (unsigned int) &outpc;
                            }
                        }
                        ad_tmp = dest_addr + var_off;

                    } else {
                        // concurrent access to tuning data via serial and CAN will cause corruption
                        chknewpage(var_blk);
                        dest_addr = (unsigned int)tables[var_blk].addrRam;
                        ck = 1;
                        ad_tmp = dest_addr + var_off;
                    }

                    if ((var_blk != 7) && ((var_off + var_byt) > tables[var_blk].n_bytes)) {
                        flagbyte3 |= flagbyte3_can_reset;
                        dest_addr = 0;
                    }

                    if (dest_addr) {

                        for (jx = 0;jx < var_byt;jx++)  {
                            if (ck && (*((unsigned char *)(ad_tmp + jx)) != *(&CAN_RB_DSR0 + jx) )) {
                                outpc.status1 |= STATUS1_NEEDBURN; // we changed some tuning data
                            }
                            *((unsigned char *)(ad_tmp + jx)) = *(&CAN_RB_DSR0 + jx);
                        }

                        if (var_blk <= 3) { // calibration data
                            /* At end of each 1k block perform a burn automatically. */
                            if ((var_off == 0x3f8) || (var_off == 0x7f8)) {
                                /* Note that this does allow misbehaving tuning software to cause a partial burn. */
                                resetholdoff = 15;
                                burn_idx = var_blk;
                                Flash_Init(); // check FDIV written to (should be at reset) likely safe?
                                flocker = 0xcc; // set semaphore to prevent burning flash thru runaway code
                                burntbl(var_off - 0x3f8); /* burn stumble is inevitable, so handling this in the interrupt is 'ok' */
                            }
                        } else if ((var_blk == 7) && (dest_addr == (unsigned int)&datax1)
                            && (var_off <= ((unsigned int)&datax1.testmodelock - (unsigned int)&datax1))
                            && ((var_off + var_byt) >= ((unsigned int)&datax1.testmodelock - (unsigned int)&datax1) + 1)) {
                            /* was the write to datax1 and covering the testmodelock.
                             (To ensure data capture doesn't disturb testmode.) */
                            /* test mode special code */
                            if (datax1.testmodelock == 12345) {
                                unsigned int ret;
                                ret = do_testmode();
                                /* no method of acting upon return value presently with protocol 0 */
                            } else {
                                flagbyte1 &= ~flagbyte1_tstmode; /* disable test mode */
                                outpc.status3 &= ~STATUS3_TESTMODE;
                            }
                        }
                    }
                }
                break;

            case MSG_REQX:  // extended
                jx = can[0].cxno_in;
                can[0].cx_msg_type[jx] = MSG_RSP;
                // destination var blk
                can[0].cx_destvarblk[jx] = CAN_RB_DSR1;
                dvar_off = ((unsigned short)CAN_RB_DSR2 << 3) |
                    ((CAN_RB_DSR3 & 0xE0) >> 5);
                can[0].cx_destvaroff[jx] = dvar_off;
                can[0].cx_dest[jx] = rcv_id;
                var_byt = CAN_RB_DSR3 & 0x1F;
                can[0].cx_varbyt[jx] = var_byt;
                var_blk = CAN_RB_DSR4;
                goto RX_REQ1;

            case MSG_REQ:  // msg to send back current value of variable(s)
                // Update related parameters for xmt ring buffer
                jx = can[0].cxno_in;
                can[0].cx_msg_type[jx] = MSG_RSP;
                // destination var blk
                can[0].cx_destvarblk[jx] = CAN_RB_DSR0;
                dvar_off = ((unsigned short)CAN_RB_DSR1 << 3) |
                    ((CAN_RB_DSR2 & 0xE0) >> 5);
                can[0].cx_destvaroff[jx] = dvar_off;
                can[0].cx_dest[jx] = rcv_id;
                var_byt = CAN_RB_DSR2 & 0x1F;
                can[0].cx_varbyt[jx] = var_byt;
RX_REQ1:;
                // put variable value(s) in xmit ring buffer
                unsigned int dest_addr = 0;
//                unsigned char save_ppage;
//                save_ppage = PPAGE;

                /* removed ability to read calibration tables - not being used */

                if ((var_blk > 3) && (var_blk < 16)) {
                    if (((var_blk != 7) && ((var_off + var_byt) > tables[var_blk].n_bytes))
                        || ((var_blk == 7) && (var_off < 0x200)  && ((var_off + var_byt) > sizeof(outpc)))
                        || ((var_blk == 7) && (var_off >= 0x200) && (var_off < 0x400) && ((var_off + var_byt) > (sizeof(datax1) + 0x200)))
                        || ((var_blk == 7) && (var_off >= 0x400) && ((var_off + var_byt) > 0x4ff))
                        ) {
                        flagbyte3 |= flagbyte3_can_reset;
                    } else {
//                        flagbyte4 |= flagbyte4_cantuning;
                        if ((var_blk == 7) && (var_off >= 0x400)) {
                            if (var_off <= 0x4ff) {
                                dest_addr = 0xf700;
                                var_off -= 0x400;
                            }
                        } else if ((var_blk == 7) && (var_off >= 0x200)) {
                            dest_addr = (unsigned int)&datax1;
                            var_off -= 0x200;
                        } else {
                            dest_addr = (unsigned int)tables[var_blk].addrRam;
                        }
                    }
                } else if ((var_blk >= 0xf0) && (var_blk <= 0xf3)) {
                    dest_addr = (unsigned int)&ram_data; // will be...
                }
                chknewpage(var_blk);

                if (dest_addr == 0xf700) {
                    unsigned int n;
                    n = conf_str_cp((unsigned char*)&can[0].cx_datbuf[jx][0], var_byt, var_off);
                    n = var_byt - n;
                    for(kx = n;kx < var_byt;kx++)  {
                        can[0].cx_datbuf[jx][kx] = 0x00; // pad with zeros
                    }

                } else if (dest_addr) {
                    unsigned int tmp_ad;
                    tmp_ad = dest_addr + var_off;
                    for(kx = 0;kx < var_byt;kx++)  {
                        can[0].cx_datbuf[jx][kx] = *((unsigned char *)(tmp_ad + kx));
                    }
                }
//                PPAGE = save_ppage;

                (void)can_inc_ring0();

                if ((var_blk >= 0xf0) && (var_blk <= 0xf3)) {
                    if (var_off == 0) {
                        flagbyte0 &= ~(FLAGBYTE0_TTHLOG | FLAGBYTE0_TRGLOG | FLAGBYTE0_COMPLOG /*| FLAGBYTE0_MAPLOG | FLAGBYTE0_MAPLOGARM*/);
                        /* ensure tooth logger off, ensure trigger logger off, ensure composite tooth logger off */
                        outpc.status3 &= ~STATUS3_DONELOG; /* clear 'complete/available' flag */
                    } else if (var_off == 1016) {
                        /* As all CAN handling is in this interrupt, we don't have a way of simply clearing the data
                          on the first read. So all tooth loggers will return one page of junk when accessed via CAN. */
                        flagbyte5 |= FLAGBYTE5_LOG_CLR; /* this actually starts the logging */
                    }
                }

                break;

            case OUTMSG_REQ:   // msg to send back current value of an outmsg 
                jx = can[0].cxno_in;
                can[0].cx_msg_type[jx] = OUTMSG_RSP;
                msg_no = CAN_RB_DSR0;   // Message number
                can[0].cx_myvarblk[jx] = msg_no;   // Use the local var block as the message number
                cur_outmsg_var[msg_no] = 0; 
                can[0].cx_destvarblk[jx] = var_blk;    // Use the message var_blk as the destination var_blk
                can[0].cx_destvaroff[jx] = var_off;    // Use the message var_off as the destination var_off
                can[0].cx_dest[jx] = rcv_id;

                // Find the number of bytes for the first response packet
                var_byt = 0;
                for (kx = 0; kx < 8; ) {
                    cur_size = flash12.outmsgs[msg_no].size[kx];
                    if ((cur_size != 0) && ((var_byt + cur_size) <= 8)) {
                        for (lx = 0; lx < cur_size; lx++) {
                            can[0].cx_datbuf[jx][var_byt] = *((unsigned char *)&outpc + flash12.outmsgs[msg_no].offset[cur_outmsg_var[msg_no]] + (unsigned int)lx);
                            var_byt++;
                        }
                        cur_outmsg_var[msg_no]++;
                    } else {
                        break;
                    }
                    kx += cur_size;
                }
                can[0].cx_varbyt[jx] = var_byt;

                (void)can_inc_ring0();
                break;

//          case OUTMSG_RSP:  // Not implemented
        		
            case MSG_XSUB:  // msg to execute a subroutine
                switch(rcv_id)  {
                    case 1:           // message to this ecu from device 1
                        can_xsub01();		// execute sub immediately here (set
                        // flag if can execute in main loop) 
                        break;
                }
                break;

            case MSG_BURN:  // msg to burn data table
                resetholdoff = 15;
                Flash_Init(); //check FDIV written to (should be at reset)
                flocker = 0xcc;
                burn_idx = var_blk;
                burntbl(0);
                outpc.status1 &= ~(STATUS1_LOSTDATA | STATUS1_NEEDBURN);
                break;
            case MSG_FWD:  // msg to forward data to the serial port
				// set up for serial xmit of the recvd CAN bytes (<= 7)
				// Unsolicited data
				if ((var_byt > 0) && (var_byt <= 7))  {
                    canbuf[16] = var_byt - 1; // serial.c will use this
					for (jx = 0;jx < canbuf[16];jx++)  {
						canbuf[jx] = *(&CAN_RB_DSR1 + jx);
					}
                    flagbyte14 |= FLAGBYTE14_SERIAL_FWD; // mainloop should send when safe
				}
                break;

            case MSG_CRC:  // msg to calc CRC32 of page - actually done in misc
                if ((var_blk == 4) || (var_blk == 5) || (var_blk == 8) || (var_blk == 9) || (var_blk == 10) || (var_blk == 11) || (var_blk == 12)) {
                    // jump to common code used by serial code
                    // if tuning remotely still need to copy flash to RAM etc.
                    chknewpage(var_blk);
                }
                crc_idx = var_blk;
                flagbyte6 |= FLAGBYTE6_CRC_CAN; // set up to do the biz from the mainloop
                // canbuf seems a fair place to store this data. Hoping nothing else tries to write it.
                canbuf[4] = CAN_RB_DSR1; // dest var blk
                canbuf[5] = rcv_id;
                break;

            case MSG_PROT:  // msg to send back current protocols
                // Update related parameters for xmt ring buffer
                jx = can[0].cxno_in;
                can[0].cx_msg_type[jx] = MSG_RSP;
                // destination var blk
                can[0].cx_destvarblk[jx] = CAN_RB_DSR1; // offset as extended
                dvar_off = ((unsigned short)CAN_RB_DSR2 << 3) |
                    ((CAN_RB_DSR3 & 0xE0) >> 5);
                can[0].cx_destvaroff[jx] = dvar_off;
                can[0].cx_dest[jx] = rcv_id;
                can[0].cx_datbuf[jx][0] = SRLVER;

                if ((CAN_RB_DSR3 & 0x1F) == 5) { // long format
                    can[0].cx_varbyt[jx] = 5;
                    *(unsigned int*)&can[0].cx_datbuf[jx][1] = SRLDATASIZE;
                    *(unsigned int*)&can[0].cx_datbuf[jx][3] = 1024; // size of ramdata
                } else { // short format
                    can[0].cx_varbyt[jx] = 1;
                }

                (void)can_inc_ring0();
                break;

            case MSG_SPND:
                if (CAN_RB_DSR1 == 1) {
                    flagbyte15 |= FLAGBYTE15_CANSUSP;
                } else {
                    flagbyte15 &= ~FLAGBYTE15_CANSUSP;
                }
                break;
        }					 // end msg_type switch
        can_status &= CLR_RCV_ERR;
    }
CAN_DONERXSW:;
    if ((CANRFLG & 0x72) != 0)  {
        // Rcv error or overrun on receive
        can_status |= RCV_ERR;
        //can_reset = 1;
    }
    /* clear RX buf full, err flags to ready for next rcv int */
    /*  (Note: can't clear err count bits) */
    CANRFLG = 0xC3;

    return;
}

