/* $Id: ms2_extra_ign.c,v 1.145 2014/02/14 15:14:19 jsmcortina Exp $
 * Copyright 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
 * James Murray and Kenneth Culver
 *
 * This file is a part of MS2/Extra.
 *
 * ign_reset()
    Origin: Al Grippo
    Major: James Murray / Kenneth Culver
    Majority: Al Grippo / James Murray / Kenneth Culver
 * ign_kill
    Origin: James Murray / Kenneth Culver
    Majority: James Murray / Kenneth Culver
 * coil_dur_table
    Origin: Al Grippo
    Majority: Al Grippo
 * wheel_fill_event_array
    Origin: Kenneth Culver
    Moderate: James Murray
    Majority: James Murray / Kenneth Culver
 * syncfirst
    Origin: Kenneth Culver
    Minor: James Murray
    Majority: Kenneth Culver
 * recheck_syncfirst
    Origin: Kenneth Culver
    Minor: Jean Belanger
    Majority: Kenneth Culver
 * do_knock
    Origin: Al Grippo
    Moderate: Add digi inputs. James Murray.
    Majority: Al Grippo / James Murray
 * SET_COIL
    Origin: James Murray / Kenneth Culver
    Majority: James Murray / Kenneth Culver
 * calc_advance
    Origin: Al Grippo
    Major: James Murray / Kenneth Culver
    Majority: Al Grippo / James Murray / Kenneth Culver
 * do_everytooth_calcs
    Origin: Kenneth Culver
    Major: James Murray / Kenneth Culver
    Majority: James Murray / Kenneth Culver
 * do_tach_mask
    Origin: Al Grippo
    Major: Large re-write. James Murray
    Majority: Al Grippo / James Murray
 *
 * You should have received a copy of the code LICENSE along with this source,
 * ask on the www.msextra.com forum if you did not.
 *
*/
#include "ms2_extra.h"

void noisefilter_lookup(void)
{
    unsigned int utmp1;

    if (flash4.NoiseFilterOpts & 0x01) {
        unsigned int noiserpm;
        if (outpc.rpm < 10) {
            noiserpm = 32767;
        } else {
            noiserpm = outpc.rpm;
        }
        utmp1 = intrp_1ditable(noiserpm,4,(unsigned int *)pg10_ptr->NoiseFilterRpm, 0,
                (unsigned int *)pg10_ptr->NoiseFilterLen);

        DISABLE_INTERRUPTS;
        NoiseFilterMin = (unsigned long)utmp1;
        ENABLE_INTERRUPTS;
    }
}

void tachmask_calc(void)
{
    // EDIS false tach masking
    if (spkmode < 2) { // EDIS only)
        if (flash4.feature4_0 & 0x20) {  // is false trig on at all ?
            unsigned long dtpred_local;

            DISABLE_INTERRUPTS;
            dtpred_local = dtpred;
            ENABLE_INTERRUPTS;

            if ((flash4.feature4_0 & 0x10)) {
                unsigned long ltmp1;
                ltmp1 = ((flash4.ICISR_pmask * dtpred_local) / 100) +
                    ((unsigned long)flash4.ICISR_tmask * 100);
                false_mask_crk = ltmp1 /192;
            } else { // simple mode
                false_mask_crk = (dtpred_local /400); // simple 48% false trig protection
            }
        } else {
            false_mask_crk = 0;
            false_mask_cam = 0;
            false_period_crk_tix = 0;
            false_period_cam_tix = 0;
        }

        // Everything else more advanced options
        // calculate false trigger hold off times for primary and secondary triggers
        // we know smallest_tooth_crk and smallest_tooth_cam
        // want to calculate false_mask_crk and false_mask_cam

    } else { // everything that isn't EDIS
        unsigned long ltmp1;
        unsigned long tmp_ticks;

        if ((flash4.NoiseFilterOpts & 0x06) || (flash4.secondtrigopts & 0x06)) { // any of the filters/masks enabled?

            if (ticks_per_deg > 25000) { // seems to be less than 1000rpm
                tmp_ticks = 25000; // rail at equiv of 1000, in case we reset/key off etc and need to regain sync
            } else {
                tmp_ticks = ticks_per_deg;
            }

            // a wheel or pseudo wheel. Use smallest tooth angle to determine mask time
            // This percentage (equiv) in here could be made variable like MS2 base's various next pulse tolerance settings

            ltmp1 = tmp_ticks * smallest_tooth_crk;
            //                    ltmp1 = ltmp1 / 2864; // (equiv to * 35%)
            ltmp1 *= (unsigned long)flash4.ICISR_pmask;
            ltmp1 /= 100000L;
            ltmp1 += (unsigned long)flash4.ICISR_tmask * 150; 

            if ((flash4.NoiseFilterOpts & 0x02) && (ltmp1 < 65535)) {
                false_period_crk_tix = (unsigned int)ltmp1;
            } else {
                false_period_crk_tix = 0;
            }

            if ((ltmp1 < 576) || (flash4.NoiseFilterOpts & 0x01) || ((flash4.NoiseFilterOpts & 0x04)==0)) { // too short OR crank noise filter OR disabled
                false_mask_crk = 0;
            } else {
                ltmp1 = ltmp1 / 192;
                false_mask_crk = (unsigned int)ltmp1;
            }

            ltmp1 = tmp_ticks * smallest_tooth_cam;
            //                    ltmp1 = ltmp1 / 2864; // (equiv to * 35% )
            ltmp1 *= (unsigned long)flash4.IC2ISR_pmask;
            ltmp1 /= 100000L;
            ltmp1 += (unsigned long)flash4.IC2ISR_tmask * 150; 

            if ((flash4.secondtrigopts & 0x02) && (ltmp1 < 65535)) {
                false_period_cam_tix = (unsigned int)ltmp1;
            } else {
                false_period_cam_tix = 0;
            }

            if ((ltmp1 < 576) || (flash4.secondtrigopts & 0x01) || ((flash4.secondtrigopts & 0x04)==0)) { // too short OR cam noise filter OR disabled
                false_mask_cam = 0;
            } else {
                ltmp1 = ltmp1 / 192;
                false_mask_cam = (unsigned int)ltmp1;
            }

        } else {
            false_mask_crk = 0;
            false_mask_cam = 0;
            false_period_crk_tix = 0;
            false_period_cam_tix = 0;
        }            
    }
}

void ign_reset(void)
{
    int a;
    // enter here at init and when in stall condition
    // Disable IC, Ign,Inj OC interrupts
    //    TIE &= 0x94; // may need review
    TIE = 0; // disable all IC/OC ints. Re-enable as required in a mo
    TSCR1 = 0x80;  // always leave timer enabled
    // reinitialize

    if (resetholdoff == 0) { // not within burn timeout period
        if (!(conf_err)) {
            PORTE &= ~0x10;   // Turn off fuel Pump.
            outpc.engine = 0;
            outpc.rpm = 0;
        }
        flagbyte2 |= flagbyte2_crank_ok;
        if ((spkmode == 2) && (flash4.spk_conf2 & SPK_CONF2_GMBYPASS)) {
            // Set HEI bypass to 0v - relies on same wiring as main spark output
            if (flash4.ICIgnOption & 0x10) { /* Going high */
#ifdef MICROSQUIRT
                PTT &= ~0x10;
#else
                PTM &= ~0x10;
#endif
            } else { /* Going low */
#ifdef MICROSQUIRT
                PTT |= 0x10;
#else
                PTM |= 0x10;
#endif
            }
        }
        tcrank_done = 0xffff;
        flagbyte2 |= flagbyte2_crank_ok;
        asecount = 0;
    }

    tooth_diff_this = 0;
    tooth_diff_last = 0;
    tooth_diff_last_1 = 0;
    tooth_diff_last_2 = 0;
    pulse_no = 0;  // This will inhibit overflow counters
    t_enable_IC = 0xFFFFFFFF;
    t_enable_IC2 = 0xFFFFFFFF;
    igncount = 0;
    altcount = 0;
    egocount = 0;
    egopstat[0] = egopstat[1] = 0;
    tpsaclk = 0;
    outpc.status3 &= ~STATUS3_CUT_FUEL;
    knk_clk = 0;
    knk_clk_test = flash4.knk_trtd;
    knk_stat = 0;
    knk_count = 0;
    outpc.knk_rtd = 0;
    spark_events = spark_events_a;
    dwell_events = dwell_events_a;
    if ((flash8.seq_inj & 0x03) != 0) {
        fuel1_events = fuel1_events_a;  
        fuel2_events = fuel2_events_a;  
        fuel3_events = fuel3_events_a;  
        fuel4_events = fuel4_events_a;  
    }       

    // Turn off all spark outputs right away
#ifdef MICROSQUIRT
    FIRE_ALL_US;
#else
    FIRE_ALL;
#endif

    // Turn Off injectors
        if (seq_inj_ctrl & SEQ_STD_INJ) {
            TCTL2 &= ~0x44;   // set outputs lo in OL1,3
            CFORC |= 0x0A;
                PWME &= ~0x14;    // turn off pwms
        } else {
#ifdef MICROSQUIRT
            TCTL2 &= ~0x54;   // set outputs lo in OL1,2,3
                TCTL1 &= ~0x01;   // set output lo in OL4
            CFORC |= 0x1E;
#else
            TCTL2 &= ~0x44;   // set outputs lo in OL1,3
                TCTL1 &= ~0xA0;   // set outputs lo in OL6,7
            CFORC |= 0xCA;
#endif
        }
    *pPTMpin[3] &= ~0x08;   // turn off inj led
    outpc.squirt = 0;       // injectors off
    IgnOCpinstate = SPK;
    // enable timer & IC interrupt
    synch = SYNC_FIRST;
    if (outpc.status1 & STATUS1_SYNCOK) { // was synced
        outpc.synccnt++;
        outpc.status1 &= ~(STATUS1_SYNCOK | STATUS1_SYNCLATCH | STATUS1_SYNCFULL);
    }
    dtpred_adder = 0;
    TFLG1 = 0xff; // clear all pending ints
    TIE |= 0x01;
    if (flagbyte5 & FLAGBYTE5_CAM) {
        TIE |= TFLG_trig2;  // enable 2nd trig ISR
        flagbyte1 &= ~flagbyte1_trig2active; // reset software flag
        flagbyte2 &= ~flagbyte2_twintrignow; // reset software flag
    }

    pulse_no = 0;
    dwl[0] = 0; // 0 means not in use
    dwl[1] = 0;
    dwl[2] = 0;
    dwl[3] = 0;
    dwl[4] = 0;
    dwl[5] = 0;
    rdwl[0] = 0; // 0 means not in use
    rdwl[1] = 0;

    trig2cnt = 0;
    tooth_no = 0;
    flagbyte3 &= ~flagbyte3_toothinit;
    flagbyte1 &= ~flagbyte1_trig2active; // clear 2nd trig
    tooth_counter = 0;
    tooth_counter_main = 0;
    syncerr = 0;
    flagbyte15 &= ~FLAGBYTE15_FIRSTRPM;
    NoiseFilterMin = 0;
    dwell_long = 0;
    dwell_us = 0;
    dwell_us2 = 0;
    last_edge = 255;
    dwellq[0].sel = 0;
    dwellq[1].sel = 0;
    nextcyl_cnt = 0;
    for (a = 0 ; a < no_triggers ; a++) {
        skipdwell[a] = 0; // don't skip this dwell
    }
    skipinj = 0;
    outpc.boostduty = 0;
    return;
}

int coil_dur_table(int delta_volt)
{
    int ix;
    long interp, interp3;
    // returns correction for coil duration vs battery voltage deviation
    //  from 12 V from table lookup in .1 ms units
    // bound input arguments
    if(delta_volt > flash4.deltV_table[NO_COILCHG_PTS-1])  {
        return((int)flash4.deltDur_table[NO_COILCHG_PTS-1]);
    }
    if(delta_volt < flash4.deltV_table[0])  {
        return((int)flash4.deltDur_table[0]);
    }
    // delta_volt in V x 10
    for(ix = NO_COILCHG_PTS - 2; ix > -1; ix--)  {
        if(delta_volt > flash4.deltV_table[ix])  {
            break;
        }
    }
    if(ix < 0)ix = 0;

    interp = flash4.deltV_table[ix + 1] - flash4.deltV_table[ix];
    if(interp != 0)  {
        interp3 = (delta_volt - flash4.deltV_table[ix]);
        interp3 = (100 * interp3);
        interp = interp3 / interp;
        if(interp < 0)interp = 0;
    }
    return((int)(flash4.deltDur_table[ix] +
                interp * (flash4.deltDur_table[ix+1] - flash4.deltDur_table[ix])/ 100));
}

void wheel_fill_event_array(ign_event *spark_events_fill, ign_event *dwell_events_fill,
                            int spk_ang, int dwl_ang, ign_time last_tooth_time,
                            unsigned int last_tooth_ang, unsigned char rotary)
{
    char start, stop, iterate, tth, wfe_err;
    int tth_ang, tmp_ang, i, j;
    ign_time spk_time, dwl_time;
    long tmptime;

    if(rotary) {
        start = 2;
        stop = 4;
    } else {
        start = 0;
        stop = no_triggers;
    }
    for (i = start; i < stop; i++) {
        if (rotary) {
            j = i - 2;
        } else {
            j = i;
        }
        wfe_err = 0;
        iterate = 0;
        tth_ang = trig_angs[j]; // trigger angle for this trigger to allow for oddfire
        tth = trigger_teeth[j];
        while (!iterate) {
            if (tth_ang > spk_ang) {  // must both be signed to get comparison to work
                // found a suitable tooth
                iterate = 1;
            } else {
                //how far do we step back in deg
                tth--;
                if (tth < 1) {
                    tth = last_tooth;
                    wfe_err++;
                    if (wfe_err > 1) {
                        iterate = 2;
                    }
                }
                tth_ang += deg_per_tooth[tth - 1]; // add on time ahead of the tooth we stepped back to
            }
        } // end while


        if (iterate == 2) {
            DISABLE_INTERRUPTS;
            asm ( "nop\n" ); // something screwed up, place for breakpoint
            ENABLE_INTERRUPTS;
            // can't continue as didn't find a valid tooth
            return;
        }

        tmp_ang = tth_ang - spk_ang; // angle after trigger tooth we found
        //convert this small angle to time
        spk_time.time_32_bits = (tmp_ang * last_tooth_time.time_32_bits) / last_tooth_ang;

        wfe_err = 0;
        while ((wfe_err < 2) && (spk_time.time_32_bits < SPK_TIME_MIN)) {
            // too soon after tooth, need to step back
            tth--;
            if (tth < 1) {
                tth = last_tooth;
                wfe_err++;
            }
            tth_ang += deg_per_tooth[tth - 1]; // add on time ahead of the tooth we stepped back to
            // recalc
            tmp_ang = tth_ang - spk_ang;
            spk_time.time_32_bits = (tmp_ang * last_tooth_time.time_32_bits) / last_tooth_ang;
        }

        if (wfe_err > 1) {
            DISABLE_INTERRUPTS;
            asm ( "nop\n" ); // something screwed up, place for breakpoint
            ENABLE_INTERRUPTS;
            // can't continue as didn't find a valid tooth
            return;
        }

        spark_events_fill[(unsigned int)i].tooth = tth;
        spark_events_fill[(unsigned int)i].time = spk_time;
        spark_events_fill[(unsigned int)i].coil = i;

        // Code limits spark advance so spark point should always be ahead of or after
        // the trigger and not crossing it.
        // During next-cyl cranking we set this flag.
        // to force a spark at trigger.
        tmptime = (spk_ang * last_tooth_time.time_32_bits) / last_tooth_ang;

        if ((flash4.dwellmode & 3) == 3) {
            //if "charge on trigger", then dwell on same tooth as scheduling spark
            //might be ok for MSD in basic dizzy mode. Might give unpredictable short pulses
            // when combined with wheel mode.
            dwell_events_fill[(unsigned int)i].tooth = spark_events_fill[(unsigned int)i].tooth;
            dwell_events_fill[(unsigned int)i].time32 = 0;
        } else {
            // do normal dwell - assume that it must always start ahead of spark
            // so start from spark tooth

            wfe_err = 0;
            iterate = 0;
            while (!iterate) {
                if (tth_ang > dwl_ang) {
                    // found a suitable tooth
                    iterate = 1;
                } else {
                    //how far do we step back in deg
                    tth--;
                    if (tth < 1) {
                        tth = last_tooth;
                        wfe_err++;
                        if (wfe_err > 1) {
                            iterate = 2;
                        }
                    }
                    tth_ang += deg_per_tooth[tth - 1]; // add on time ahead of the tooth we stepped back to
                }
            } // end while

            if (iterate == 2) {
                DISABLE_INTERRUPTS;
                asm ( "nop\n" ); // something screwed up, place for breakpoint
                ENABLE_INTERRUPTS;
                // can't continue as didn't find a valid tooth
                return;
            }

            tmp_ang = tth_ang - dwl_ang; // angle after trigger tooth we found
            //convert this small angle to time
            dwl_time.time_32_bits = (tmp_ang * last_tooth_time.time_32_bits) / last_tooth_ang;

            wfe_err = 0;
            while ((wfe_err < 2) && (dwl_time.time_32_bits < SPK_TIME_MIN)) {
                // too soon after tooth, need to step back
                tth--;
                if (tth < 1) {
                    tth = last_tooth;
                    wfe_err++;
                }
                tth_ang += deg_per_tooth[tth - 1]; // add on time ahead of the tooth we stepped back to
                // recalc
                tmp_ang = tth_ang - dwl_ang;
                dwl_time.time_32_bits = (tmp_ang * last_tooth_time.time_32_bits) / last_tooth_ang;
            }

            if (wfe_err > 1) {
                DISABLE_INTERRUPTS;
                asm ( "nop\n" ); // something screwed up, place for breakpoint
                ENABLE_INTERRUPTS;
                // can't continue as didn't find a valid tooth
                return;
            }

            dwell_events_fill[(unsigned int)i].tooth = tth;
            dwell_events_fill[(unsigned int)i].time = dwl_time;
        }
        dwell_events_fill[(unsigned int)i].coil = i;

        // if dwell has stepped back then force dwell on tooth just before spark
        // or if dwell is on same tooth but fairly short then force it
        if ((dwell_events_fill[(unsigned int)i].tooth != spark_events_fill[(unsigned int)i].tooth)
                || ( (dwell_events_fill[(unsigned int)i].tooth == spark_events_fill[(unsigned int)i].tooth)
                    && (dwell_events_fill[(unsigned int)i].time32 < 170)
                    && (dwell_events_fill[(unsigned int)i].time32 >= 0) ) ) {
            spark_events_fill[(unsigned int)i].ftooth = spark_events_fill[(unsigned int)i].tooth;
        } else {
            spark_events_fill[(unsigned int)i].ftooth = 0;
        }
    }
}

void wheel_fill_fuel_event_array(fuel_event *fuel_events_fill, int fuel_ang, ign_time last_tooth_time,
                                 unsigned int last_tooth_ang, unsigned char siamese)
{
    char start, stop, iterate, tth, wfe_err;
    int tth_ang, tmp_ang;
    unsigned int fuel_time;
    unsigned char i, j, increment, pos=0;

    if(siamese == 4) {
        // Semi-sequential siamese: trigger on cylinder 1 (and 4)
        start = 0;
        stop = no_triggers;
        increment = 2;
    } else if (siamese < 4) {
        start = siamese;
        stop = siamese + 1;
        increment = 1;
        pos = 1 - (siamese / 2);
    } else {
        start = siamese - 10;
        stop = no_triggers;
        increment = no_triggers / no_squirts;
    }
    for (i = start, j = pos; i < stop; i+=increment, j++) {
        int tmp_inj_ang;

        wfe_err = 0;
        iterate = 0;
        tth_ang = trig_angs[i]; // trigger angle for this trigger to allow for oddfire
        tth = trigger_teeth[i];
        tmp_inj_ang = fuel_ang;

        if (tth_ang > tmp_inj_ang) {
            tmp_inj_ang += cycle_deg;
        }

        while (!iterate) {
            if (tth_ang > tmp_inj_ang) {  // must both be signed to get comparison to work
                // found a suitable tooth
                iterate = 1;
            } else {
                //how far do we step back in deg
                tth--;
                if (tth < 1) {
                    tth = last_tooth;
                    wfe_err++;
                    if (wfe_err > 2) {
                        iterate = 3;
                    }
                }
                tth_ang += deg_per_tooth[tth - 1]; // add on time ahead of the tooth we stepped back to
            }
        } // end while


        if (iterate == 3) {
            DISABLE_INTERRUPTS;
            asm ( "nop\n" ); // something screwed up, place for breakpoint
            ENABLE_INTERRUPTS;
            // can't continue as didn't find a valid tooth
            return;
        }

        tmp_ang = tth_ang - tmp_inj_ang; // angle after trigger tooth we found
        //convert this small angle to software timer count
        fuel_time = (unsigned int)(((tmp_ang * last_tooth_time.time_32_bits) / last_tooth_ang) / 192L);

        if (fuel_time < 1) {
            fuel_time = 1;
        }
        fuel_events_fill[j].tooth = tth;
        fuel_events_fill[j].time = fuel_time;
        fuel_events_fill[j].inj = j;
    }
}

void syncfirst(void)
{
    signed int tmp_tooth;
    unsigned char i, start_val;

    // convenient common place to put this code
    if (((flash4.spk_mode3 & 0xe0) == 0x60) // this is permanent wasted-COP mode
        && (!((flash4.spk_mode3 & 0x01) // but not the things needed for MAP phase detection
            && (flash4.no_cyl <= 2) && ((flash4.spk_config & 0x0c) == 0x0c)
            && ((flash4.spk_mode3 & 0xe0) == 0x60)
            && ((flash4.mapsample_opt & 0x04) == 0)))) {
        synch |= SYNC_WCOP2;
    }

/* check if we are somehow already at high rpms
   in that case skip these checks as not enough CPU time
    was previously preventing re-sync at higher rpms */
    if ((tooth_diff_this && (tooth_diff_this < 1800)) || (no_triggers == 1)) { // was 750
        start_val = 255;
    } else {
        start_val = 0;
    }

/* find next dwell event so we spark ASAP */
    fuel_cntr = start_val;
    tmp_tooth = tooth_no;
    while (!fuel_cntr) {
        tmp_tooth++;
        if (tmp_tooth > last_tooth) {
            tmp_tooth = 0;
        }
        if (tmp_tooth == tooth_no) {
            // been all the way around without finding anything - oops!
            fuel_cntr = 255;
        }
        for (i = 0; i < no_triggers; i++) {
            if (dwell_events[i].tooth == tmp_tooth) {
                fuel_cntr = i;
            } 
        }
    }
    if (fuel_cntr == 255) {
        fuel_cntr = 0;
    }

    next_spark.time = spark_events[fuel_cntr].time;
    next_spark.tooth = spark_events[fuel_cntr].tooth;
    next_spark.coil = spark_events[fuel_cntr].coil;
    next_spark.ftooth = spark_events[fuel_cntr].ftooth;
    next_spark.fs = spark_events[fuel_cntr].fs;
    next_dwell.time = dwell_events[fuel_cntr].time;
    next_dwell.tooth = dwell_events[fuel_cntr].tooth;
    next_dwell.coil = dwell_events[fuel_cntr].coil;

/* find next map event so we capture map ASAP */
    fuel_cntr = start_val;
    tmp_tooth = tooth_no;
    while (!fuel_cntr) {
        tmp_tooth++;
        if (tmp_tooth > last_tooth) {
            tmp_tooth = 0;
        }
        if (tmp_tooth == tooth_no) {
            // been all the way around without finding anything - oops!
            fuel_cntr = 255;
        }
        for (i = 0; i < no_triggers; i++) {
            if (map_start_event[i].tooth == tmp_tooth) {
                fuel_cntr = i;
            } 
        }
    }
    if (fuel_cntr == 255) {
        fuel_cntr = 0;
    }

    next_map_start_event.time = map_start_event[fuel_cntr].time;
    next_map_start_event.tooth = map_start_event[fuel_cntr].tooth;
    next_map_start_event.evnum = map_start_event[fuel_cntr].evnum;
    // don't use ftooth or fs in dwell array
        if ((flash8.seq_inj & 0x03) != 0) {     
                unsigned char inj_event;
                if ((flash8.seq_inj & 0x03) > 1) {      
                        inj_event = fuel_cntr>>1;
                } else {
                        inj_event = fuel_cntr / no_inj;
                }
            next_fuel1_event.time = fuel1_events[inj_event].time;
            next_fuel1_event.tooth = fuel1_events[inj_event].tooth;
            next_fuel1_event.inj = fuel1_events[inj_event].inj;
            next_fuel2_event.time = fuel2_events[inj_event].time;
            next_fuel2_event.tooth = fuel2_events[inj_event].tooth;
            next_fuel2_event.inj = fuel2_events[inj_event].inj;
                if (!(seq_inj_ctrl & SEQ_STD_INJ)) {
                    next_fuel3_event.time = fuel3_events[inj_event].time;
                    next_fuel3_event.tooth = fuel3_events[inj_event].tooth;
                    next_fuel3_event.inj = fuel3_events[inj_event].inj;
                    next_fuel4_event.time = fuel4_events[inj_event].time;
                    next_fuel4_event.tooth = fuel4_events[inj_event].tooth;
                    next_fuel4_event.inj = fuel4_events[inj_event].inj;
                }
        }

/* find next fuel event */
    fuel_cntr = start_val;
    tmp_tooth = tooth_no;
    while (!fuel_cntr) {
        tmp_tooth++;
        if (tmp_tooth > last_tooth) {
            tmp_tooth = 0;
        }
        if (tmp_tooth == tooth_no) {
            // been all the way around without finding anything - oops!
            fuel_cntr = 255;
        }
        for (i = 0; i < no_triggers; i++) {
            if (trigger_teeth[i] == tmp_tooth) {
                fuel_cntr = i;
            } 
        }
    }
    if (fuel_cntr == 255) {
        fuel_cntr = 0;
    }
    next_fuel = trigger_teeth[fuel_cntr];

    if ((next_spark.tooth != 0) &&
            (next_dwell.tooth != 0)) {
        synch &= ~SYNC_FIRST;  // must have non zero values before declaring sync not first
    }

    if (((spkmode == 4) || (spkmode == 7)) && ((flash4.EngStroke & 0x03) == 0x03)) {
        next_spk_trl.time = spark_events[2].time;
        next_spk_trl.tooth = spark_events[2].tooth;
        next_spk_trl.coil = spark_events[2].coil;
        next_dwl_trl.time = dwell_events[2].time;
        next_dwl_trl.tooth = dwell_events[2].tooth;
        next_dwl_trl.coil = dwell_events[2].coil;   
        if ((next_spk_trl.tooth != 0) && (next_dwl_trl.tooth != 0)) {
            synch &= ~SYNC_FIRST;  // must have non zero values before declaring sync not first
        }
    }
    outpc.status1 |= STATUS1_SYNCOK;
}

void recheck_syncfirst(void)
{
    /* Check for situation where SYNC_FIRST is off, but trigger teeth are not
        actually valid. This can occur if MS2 is powered on (or a reset) at high
        rpms and syncfirst runs before the event arrays are filled. More prevalent
        with certain tooth#1 angles.
        Does not happen during a normal start.
    */
    if (!(synch & SYNC_FIRST)) { // supposed to have picked up next trigger points
        if (next_spark.tooth == 0) { // but not actually pointing to valid tooth
            DISABLE_INTERRUPTS;
            next_spark.time = spark_events[0].time;
            next_spark.tooth = spark_events[0].tooth;
            next_spark.coil = spark_events[0].coil;
            next_spark.ftooth = spark_events[0].ftooth;
            next_spark.fs = spark_events[0].fs;
            ENABLE_INTERRUPTS;
        }
        if (next_dwell.tooth == 0) {
            DISABLE_INTERRUPTS;
            next_dwell.time = dwell_events[0].time;
            next_dwell.tooth = dwell_events[0].tooth;
            next_dwell.coil = dwell_events[0].coil;
            ENABLE_INTERRUPTS;
        }
        if (next_map_start_event.tooth == 0) {
            DISABLE_INTERRUPTS;
            next_map_start_event.time = map_start_event[fuel_cntr].time;
            next_map_start_event.tooth = map_start_event[fuel_cntr].tooth;
            next_map_start_event.evnum = map_start_event[fuel_cntr].evnum;
            ENABLE_INTERRUPTS;
        }
        if ((flash8.seq_inj & 0x03) != 0) {
            if (next_fuel1_event.tooth == 0) {
                DISABLE_INTERRUPTS;
                next_fuel1_event.time = fuel1_events[0].time;
                next_fuel1_event.tooth = fuel1_events[0].tooth;
                next_fuel1_event.inj = fuel1_events[0].inj;
                ENABLE_INTERRUPTS;
            }
            if (next_fuel2_event.tooth == 0) {
                DISABLE_INTERRUPTS;
                next_fuel2_event.time = fuel2_events[0].time;
                next_fuel2_event.tooth = fuel2_events[0].tooth;
                next_fuel2_event.inj = fuel2_events[0].inj;
                ENABLE_INTERRUPTS;
            }
            if (!(seq_inj_ctrl & SEQ_STD_INJ)) {
                if (next_fuel3_event.tooth == 0) {
                    DISABLE_INTERRUPTS;
                    next_fuel3_event.time = fuel3_events[0].time;
                    next_fuel3_event.tooth = fuel3_events[0].tooth;
                    next_fuel3_event.inj = fuel3_events[0].inj;
                    ENABLE_INTERRUPTS;
                }
                if (next_fuel4_event.tooth == 0) {
                    DISABLE_INTERRUPTS;
                    next_fuel4_event.time = fuel4_events[0].time;
                    next_fuel4_event.tooth = fuel4_events[0].tooth;
                    next_fuel4_event.inj = fuel4_events[0].inj;
                    ENABLE_INTERRUPTS;
                }
            }
        }
        if (next_fuel == 0) {
            DISABLE_INTERRUPTS;
            next_fuel = trigger_teeth[fuel_cntr];
            ENABLE_INTERRUPTS;
        }
        if (((spkmode == 4) || (spkmode == 7)) && ((flash4.EngStroke & 0x03) == 0x03)) {
            if (next_spk_trl.tooth == 0) {
                DISABLE_INTERRUPTS;
                next_spk_trl.time = spark_events[2].time;
                next_spk_trl.tooth = spark_events[2].tooth;
                next_spk_trl.coil = spark_events[2].coil;
                ENABLE_INTERRUPTS;
            }
            if (next_dwl_trl.tooth == 0) {
                DISABLE_INTERRUPTS;
                next_dwl_trl.time = dwell_events[2].time;
                next_dwl_trl.tooth = dwell_events[2].tooth;
                next_dwl_trl.coil = dwell_events[2].coil;
                ENABLE_INTERRUPTS;
            }
        }
    }
}

void SET_COIL(unsigned char *coil, unsigned char *bits) {
    if ((flash4.EngStroke & 0x03) == 0x03) {
        if (*coil == TRIGA) {
            *bits |= COILABIT;
        } else if (*coil == TRIGB) {
            *bits |= COILBBIT;
        } else if (*coil == TRIGC) {
            *bits |= COILCBIT;
        } else if (*coil == TRIGD) {
            *bits |= COILDBIT;
        }
    } else {
        if (*coil == TRIGA) {
            *bits |= COILABIT;
        } else if (*coil == TRIGB) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else {
                *bits |= COILBBIT;
            }
        } else if (*coil == TRIGC) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILABIT;
            } else {
                *bits |= COILCBIT;
            }
        } else if (*coil == TRIGD) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILBBIT;
            } else if (num_spk == 3) {
                *bits |= COILABIT;
            } else {
                *bits |= COILDBIT;
            }
        } else if (*coil == TRIGE) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILABIT;
            } else if (num_spk == 3) {
                *bits |= COILBBIT;
            } else if (num_spk == 4) {
                *bits |= COILABIT;
            } else {
                *bits |= COILEBIT;
            }
        } else if (*coil == TRIGF) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILBBIT;
            } else if (num_spk == 3) {
                *bits |= COILCBIT;
            } else if (num_spk == 4) {
                *bits |= COILBBIT;
            } else {
                *bits |= COILFBIT;
            }
        } else if (*coil == TRIGG) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILABIT;
            } else if (num_spk == 4) {
                *bits |= COILCBIT;
            } else {
                *bits |= COILGBIT;
            }
        } else if (*coil == TRIGH) {
            if (num_spk == 1) {
                *bits |= COILABIT;
            } else if (num_spk == 2) {
                *bits |= COILBBIT;
            } else if (num_spk == 4) {
                *bits |= COILDBIT;
            } else {
                *bits |= COILHBIT;
            }
        }
    }
}

void calc_advance(long *lsum)
{
    unsigned char uctmp;
    int base_timing, adv1 = 0, adv2 = 0, adv3 = 0, rl_rtd = 0, idlecor = 0;

    //DIST_ADV:
    // Table switching
    uctmp = 0; //uctmp is local tmp var
    // nitrous control will also use this var as pin behaviour will be different
    //note: doesn't block DT as not believed to have any effect
    if (flash5.feature5_0 & 0x10) { //  table switching
        if ((flash5.feature5_0 & 0x60) == 0) {
            if (flash5.ts_port & 0x80) {
                // Remote switch
                if (!(outpc.gpioport[2] & (0x01 << ((flash5.ts_port & 0x70) >> 4)))) {
                    uctmp = 1;
                }
            } else if ((flash5.ts_port & 0x0f) == 1) {
                // Local switch
                if (!(PORTE & 0x1)) { // PE0/FLEX. Pullup always enabled.
                    uctmp = 1;
                }
            } else {
                // Local switch
                if (!(PORTE & 0x2)) { // the bastard pin. Pullup always enabled.
                    uctmp = 1;
                }
            }
        } else if ((flash5.feature5_0 & 0x60) == 0x20) {
            if (outpc.rpm > flash5.tss_rpm) {
                uctmp = 1;
            }
        } else if ((flash5.feature5_0 & 0x60) == 0x40) {
            if (outpc.map > flash5.tss_kpa) {
                uctmp = 1;
            }
        } else if ((flash5.feature5_0 & 0x60) == 0x60) {
            if (outpc.tps > flash5.tss_tps) {
                uctmp = 1;
            }
        }
    }

    if (uctmp) {
        outpc.status1 |= status1_stblsw;
    } else {
        outpc.status1 &= ~status1_stblsw;
    }

    // Calculate ignition advance
    if (outpc.engine & ENGINE_CRANK) {
        if ( ((spkmode == 2) || (spkmode == 3) || (spkmode == 14) )&& (flash4.adv_offset <= 200) ) {  //2,3
            *lsum = flash4.adv_offset; // next-cyl fire at trigger on in dizzy modes
        } else {
            *lsum = flash4.crank_timing;
        }
        base_timing = *lsum;
    } else if (pg4_ptr->timing_flags & TIMING_FIXED) {
        base_timing = pg4_ptr->fixed_timing;
        *lsum = pg4_ptr->fixed_timing;
    } else {

        if (uctmp) { // table switching
            adv3 = intrp_2ditable(outpc.rpm, outpc.ignload, NO_SRPMS, NO_SMAPS,
                    &pg8_ptr->srpm_table3[0], &pg8_ptr->smap_table3[0], (int *)&pg8_ptr->adv_table3[0][0]);
            *lsum = adv3;
        } else {
            adv1 = intrp_2ditable(outpc.rpm, outpc.ignload, NO_SRPMS, NO_SMAPS,
                    &pg10_ptr->srpm_table[0][0], &pg10_ptr->smap_table[0][0], (int *)&pg10_ptr->adv_table[0][0][0]);
            *lsum = adv1;
        }
        base_timing = *lsum;

        if (flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_LOAD 
                || flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_RPM
                || flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_ADAPTIVE) {
            char doit = 0;
            if ((flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_CL_COND) 
                && (outpc.status2 & status2_pwmidle_closedloop)) {
                doit = 1;
            } else if (!(flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_CL_COND) &&
            (outpc.tps < flash8.idleadvance_tps) &&
                    (outpc.rpm < flash8.idleadvance_rpm) &&
                    (outpc.fuelload > flash8.idleadvance_load) &&
                    (outpc.clt > flash8.idleadvance_clt)) {
                doit = 1;
            } else {
                flagbyte4 &= ~flagbyte4_idleadvreset;
            }

            if(doit && (idle_advance_timer >= flash8.idleadvance_delay) &&
                        (flagbyte4 & flagbyte4_idleadvreset)) {
                    /* deg x 10 */

                    long t = 0;

                    if (flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_LOAD) {
                        t = intrp_1ditable(outpc.fuelload, 5, (int *)pg12_ptr->idleadvance_loads, 1, (int *)pg12_ptr->idleadvance_curve);
                    }

                    if (flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_RPM) {
                        t = intrp_1ditable(outpc.rpm, 5, (int *)pg12_ptr->idleadvance_rpms, 1, (int *)pg12_ptr->idleadvance_curve);
                    }

                    if (flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_ADAPTIVE) {
                        t = intrp_1ditable(outpc.rpm - outpc.cl_idle_targ_rpm, 5, (int *)pg12_ptr->idleadvance_rpms, 1, (int *)pg12_ptr->idleadvance_curve);
                    }

                    if (flash8.idle_special_ops & IDLE_SPECIAL_OPS_IDLEADVANCE_ADDER) {
                        idlecor = t;
                        *lsum += t;
                    } else {
                        base_timing = t;
                        *lsum = t;
                    }
                } else {
                    if (!(flagbyte4 &flagbyte4_idleadvreset)) {
                        DISABLE_INTERRUPTS;
                        idle_advance_timer = 0;
                        ENABLE_INTERRUPTS;
                        flagbyte4 |= flagbyte4_idleadvreset;
                    }
                }
        }

        *lsum += outpc.cold_adv_deg - outpc.knk_rtd; // degx 10
        // Subtract retard vs manifold air temp
        *lsum -= intrp_1dctable(outpc.airtemp, NO_MAT_TEMPS, (int *)flash5.MatTemps, 0, (unsigned char *)flash5.MatSpkRtd);    // %
        // correct for flex fuel;
        *lsum += outpc.flex_advance;    // degx10
        /* If the second table is enabled, use it */
        if (flash4.IgnAlgorithm & 0xF0) {
            adv2 = intrp_2ditable(outpc.rpm, outpc.ignload2, NO_SRPMS, NO_SMAPS,
                    &pg10_ptr->srpm_table[1][0], &pg10_ptr->smap_table[1][0], (int *)&pg10_ptr->adv_table[1][0][0]);
            /* Ign only additive */
            *lsum += adv2;
        }
        // rev limit
        if (flash4.RevLimOption & 1)  {
            if (outpc.rpm > RevLimRpm2) {
                rl_rtd = flash4.RevLimMaxRtd;                  // deg x 10
                outpc.status3 |= STATUS3_REVLIMSFT;
            } else if (outpc.rpm > RevLimRpm1)  {
                rl_rtd = (((long)flash4.RevLimMaxRtd*(outpc.rpm- RevLimRpm1))/
                        (RevLimRpm2 - RevLimRpm1));             // deg x 10
                outpc.status3 |= STATUS3_REVLIMSFT;
            } else {
                outpc.status3 &= ~STATUS3_REVLIMSFT;
            }
            *lsum -= rl_rtd;     // deg x 10
        } else if (flash4.RevLimOption & 2) {
            if ((outpc.rpm > RevLimRpm1) && (*lsum > flash4.RevLimRtdAng)) {      // made sure we are going to retard.
                rl_rtd = *lsum - flash4.RevLimRtdAng; // calculate in retard terms
                *lsum = flash4.RevLimRtdAng;
                outpc.status3 |= STATUS3_REVLIMSFT;
            } else {
                outpc.status3 &= ~STATUS3_REVLIMSFT;
            }

        }
    }
    outpc.adv1 = adv1;
    outpc.adv2 = adv2;
    outpc.adv3 = adv3;
    outpc.base_advance = base_timing;
    outpc.idle_cor_advance = idlecor;
    outpc.revlim_retard = rl_rtd;
}
